# coding: utf-8
"""
contains utils for inputs and outputs
"""
from __future__ import annotations


import contextlib

import h5py
from silx.io.url import DataUrl
from silx.io.utils import open as open_hdf5


class _BaseReader(contextlib.AbstractContextManager):
    def __init__(self, url: DataUrl):
        if not isinstance(url, DataUrl):
            raise TypeError("url should be an instance of DataUrl")
        if url.scheme() not in ("silx", "h5py"):
            raise ValueError("Valid scheme are silx and h5py")
        if url.data_slice() is not None:
            raise ValueError(
                "Data slices are not managed. Data path should "
                "point to a bliss node (h5py.Group)"
            )
        self._url = url
        self._file_handler = None

    def __exit__(self, *exc):
        return self._file_handler.close()


class EntryReader(_BaseReader):
    """Context manager used to read a bliss node"""

    def __enter__(self):
        self._file_handler = open_hdf5(filename=self._url.file_path())
        entry = self._file_handler[self._url.data_path()]
        if not isinstance(entry, h5py.Group):
            raise ValueError("Data path should point to a bliss node (h5py.Group)")
        return entry


class DatasetReader(_BaseReader):
    """Context manager used to read a bliss node"""

    def __enter__(self):
        self._file_handler = open_hdf5(filename=self._url.file_path())
        entry = self._file_handler[self._url.data_path()]
        if not isinstance(entry, h5py.Dataset):
            raise ValueError("Data path should point to a dtaset (h5py.Dataset)")
        return entry



# coding: utf-8
from __future__ import annotations

import pytest
from silx.gui.utils.testutils import SignalListener

from tomwer.gui.control.datalistener import ConfigurationWidget
from tomwer.gui.utils.qt_utils import block_signals
from tomwer.tests.utils import skip_gui_test
from tomwer.tests.conftest import qtapp  # noqa F401


@pytest.mark.skipif(skip_gui_test(), reason="skip gui test")
def test_TestDataListenerConfiguration(
    qtapp,  # noqa F811
):
    configWidget = ConfigurationWidget(parent=None, enable_host_pinging=False)
    sig_listener = SignalListener()
    configWidget.sigConfigurationChanged.connect(sig_listener)
    with block_signals(configWidget):
        configWidget.setHost("localhost")

    assert configWidget.getConfiguration() == {"host": "localhost", "port": 4000}
    with block_signals(configWidget):
        configWidget.setPort(0)
    with block_signals(configWidget):
        configWidget.setHost("toto")
    assert configWidget.getConfiguration() == {"host": "toto", "port": 0}
    assert sig_listener.callCount() == 0



from tomwer.io.utils.raw_and_processed_data import file_is_on_processed_data


def test_file_is_on_processed_data():
    """test file_is_on_processed_data function"""
    assert file_is_on_processed_data("/path/to/file") == False
    assert file_is_on_processed_data("/path/to/PROCESSED_DATA/file") == True
    assert file_is_on_processed_data("/path/to/RAW_DATA/file") == False
    assert file_is_on_processed_data("/path/to/RAW_DATA/PROCESSED_DATA/file") == True
    assert file_is_on_processed_data("/path/to/PROCESSED_DATA/RAW_DATA/file") == False



from __future__ import annotations

import os

from nxtomomill.converter.hdf5.utils import PROCESSED_DATA_DIR_NAME, RAW_DATA_DIR_NAME

from tomwer.io.utils.raw_and_processed_data import to_processed_data_path
from tomwer.core.scan.scanbase import TomwerScanBase
from .output import DATASET_GALLERY_DIR_NAME, PROPOSAL_GALLERY_DIR_NAME

__all__ = ["deduce_dataset_gallery_location", "deduce_proposal_GALLERY_location"]


def deduce_dataset_gallery_location(scan_obj: TomwerScanBase) -> str:
    """
    From scan path deduce the 'dataset' path to the gallery.
    Warning: dataset gallery is different then the 'proposal' gallery
    """
    if not isinstance(scan_obj, TomwerScanBase):
        raise TypeError(f"'scan_obj' is expected to be an instance of {TomwerScanBase}")

    file_path = os.path.abspath(scan_obj.path)

    split_path = file_path.split(os.sep)
    # reverse it to find the lower level value of 'PROCESSED_DATA_DIR_NAME' or 'RAW_DATA_DIR_NAME' if by any 'chance' has several in the path
    # then we will replace the 'lower one' in the string. This is where the GALLERY will be added
    split_path = split_path[::-1]
    # check if already contain in a "PROCESSED_DATA" directory
    try:
        index_processed_data = split_path.index(PROCESSED_DATA_DIR_NAME)
    except ValueError:
        pass
        index_processed_data = None
    try:
        index_raw_data = split_path.index(RAW_DATA_DIR_NAME)
    except ValueError:
        # if the value is not in the list
        index_raw_data = None

    if index_processed_data is None and index_raw_data is None:
        # if not in any "PROCESSED_DATA" or "RAW_DATA" directory
        return scan_obj.get_relative_file(
            file_name=DATASET_GALLERY_DIR_NAME, with_dataset_prefix=False
        )
    elif index_processed_data is not None and index_raw_data is not None:
        if index_raw_data > index_processed_data:
            # if PROCESSED_DATA lower in the path than RAW_DATA
            split_path[index_processed_data] = RAW_DATA_DIR_NAME
    # reorder path to original
    split_path = list(split_path[::-1])
    split_path.append(DATASET_GALLERY_DIR_NAME)
    # move it to PROCESSED_DATA when possible
    path = os.sep.join(split_path)
    path = to_processed_data_path(path)
    return path


def deduce_proposal_GALLERY_location(scan_obj: TomwerScanBase) -> str:
    """
    Policy: look if the scan_obj.path is in 'PROCESSED_DATA_DIR_NAME' or 'RAW_DATA_DIR_NAME' directories.
    If find any (before any 'GALLERY_DIR_NAME' directory) replace it "GALLERY_DIR_NAME".
    If none of those are found then create it at the same level as the scan

    :param scan_obj: scan_obj for which we want the GALLERY directory
    :return: gallery path (to save screenshots for example)
    """
    if not isinstance(scan_obj, TomwerScanBase):
        raise TypeError(f"'scan_obj' is expected to be an instance of {TomwerScanBase}")

    file_path = os.path.abspath(scan_obj.path)

    split_path = file_path.split(os.sep)
    # reverse it to find the lower level value of 'PROCESSED_DATA_DIR_NAME' or 'RAW_DATA_DIR_NAME' if by any 'chance' has several in the path
    # then we will replace the 'lower one' in the string. This is where the GALLERY will be added
    split_path = split_path[::-1]
    # check if already contain in a "PROCESSED_DATA" directory
    try:
        index_processed_data = split_path.index(PROCESSED_DATA_DIR_NAME)
    except ValueError:
        pass
        index_processed_data = None
    try:
        index_raw_data = split_path.index(RAW_DATA_DIR_NAME)
    except ValueError:
        # if the value is not in the list
        index_raw_data = None

    if index_processed_data is None and index_raw_data is None:
        # if not in any "PROCESSED_DATA" or "RAW_DATA" directory
        return scan_obj.get_relative_file(
            file_name=PROPOSAL_GALLERY_DIR_NAME, with_dataset_prefix=False
        )
    elif index_processed_data is not None and index_raw_data is not None:
        if index_raw_data > index_processed_data:
            # if PROCESSED_DATA lower in the path than RAW_DATA
            split_path[index_processed_data] = PROPOSAL_GALLERY_DIR_NAME
        else:
            # if RAW_DATA lower in the path than PROCESSED_DATA
            split_path[index_raw_data] = PROPOSAL_GALLERY_DIR_NAME
    elif index_raw_data is not None:
        # if the path contains only PROCESSED_DATA or RAW_DATA (expected behavior for online acquistion)
        split_path[index_raw_data] = PROPOSAL_GALLERY_DIR_NAME
    else:
        assert index_processed_data is not None, "index_processed_data is None"
        split_path[index_processed_data] = PROPOSAL_GALLERY_DIR_NAME

    # reorder path to original
    split_path = split_path[::-1]
    return os.sep.join(split_path)



"""contain the AxisTask class"""

from __future__ import annotations

import logging

from nabu.pipeline.estimators import estimate_cor
from nabu.resources.nxflatfield import update_dataset_info_flats_darks
from processview.core.manager import DatasetState, ProcessManager
from processview.core.superviseprocess import SuperviseProcess

import tomwer.version
from tomwer.core.process.reconstruction.utils.cor import absolute_pos_to_relative
from tomwer.core.process.task import Task
from tomwer.core.scan.nxtomoscan import NXtomoScan
from tomwer.core.scan.edfscan import EDFTomoScan
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.scan.scanfactory import ScanFactory
from tomwer.core.utils import image, logconfig
from tomwer.core.utils.scanutils import data_identifier_to_scan
from tomwer.utils import docstring

from .mode import AxisMode
from .params import (
    DEFAULT_CMP_N_SUBSAMPLING_Y,
    DEFAULT_CMP_NEAR_POS,
    DEFAULT_CMP_NEAR_WIDTH,
    DEFAULT_CMP_OVERSAMPLING,
    DEFAULT_CMP_TAKE_LOG,
    DEFAULT_CMP_THETA,
    AxisRP,
)
from .projectiontype import ProjectionType

_logger = logging.getLogger(__name__)

# vertically, work on a window having only a percentage of the frame.
pc_height = 10.0 / 100.0
# horizontally. Global method supposes the COR is more or less in center
# % of the detector:

pc_width = 50.0 / 100.0


def _absolute_pos_to_relative_with_warning(absolute_pos: float, det_width: int | None):
    """
    nabu returns the value as absolute. tomwer needs it as relative
    Also handle the case (unlikely) the detector width cannot be found
    """
    if det_width is None:
        det_width = 2048
        _logger.warning("unable to find image width. Set width to 2048")
    else:
        det_width = det_width
    return absolute_pos_to_relative(absolute_pos=absolute_pos, det_width=det_width)


def adapt_tomwer_scan_to_nabu(scan: TomwerScanBase, do_flatfield: bool):
    """simple util to convert tomwer scan to a nabu DataAnalizer and
    updating infos regarding flat and dark if needed
    """
    dataset_infos = scan.to_nabu_dataset_analyser()
    if isinstance(scan, NXtomoScan):
        try:
            update_dataset_info_flats_darks(
                dataset_infos,
                flatfield_mode=do_flatfield,
                loading_mode="load_if_present",
            )
        except ValueError as exception:
            # nabu raise an error if no darks / flats set. But this can make sense at this stage if the NXtomo has no
            # raw dark / flat and is already normalized. In this case only fire a warning
            if (
                scan.reduced_darks is not None
                and len(scan.reduced_darks) > 0
                and scan.reduced_flats is not None
                and len(scan.reduced_flats) > 0
            ):
                raise exception
            else:
                _logger.warning(
                    "Fail to update nabu dataset info flats and darks. Expected if the dataset contains already normalized projections"
                )

    return dataset_infos


def get_composite_options(scan) -> tuple:
    theta = scan.axis_params.composite_options.get("theta", DEFAULT_CMP_THETA)
    n_subsampling_y = scan.axis_params.composite_options.get(
        "n_subsampling_y", DEFAULT_CMP_N_SUBSAMPLING_Y
    )
    oversampling = scan.axis_params.composite_options.get(
        "oversampling", DEFAULT_CMP_OVERSAMPLING
    )
    take_log = scan.axis_params.composite_options.get("take_log", DEFAULT_CMP_TAKE_LOG)

    near_pos = scan.axis_params.composite_options.get("near_pos", DEFAULT_CMP_NEAR_POS)
    near_width = scan.axis_params.composite_options.get(
        "near_width", DEFAULT_CMP_NEAR_WIDTH
    )
    return theta, n_subsampling_y, oversampling, take_log, near_pos, near_width


def read_scan_x_rotation_axis_pixel_position(scan: TomwerScanBase):
    """read center of rotation estimation from metadata"""
    if isinstance(scan, EDFTomoScan):
        raise TypeError("EDFTomoScan have not information about estimated cor position")
    else:
        res = scan.x_rotation_axis_pixel_position
        if res is None:
            _logger.error(
                f"unable to find x_rotation_axis_pixel_position from {scan.get_identifier()}"
            )
    return res


class NoAxisUrl(Exception):
    pass


class AxisTask(
    Task,
    SuperviseProcess,
    input_names=("data",),
    output_names=("data",),
    optional_input_names=("serialize_output_data",),
):
    """
    Process used to compute the center of rotation of a scan
    """

    def __init__(
        self,
        process_id=None,
        varinfo=None,
        inputs=None,
        node_id=None,
        node_attrs=None,
        execinfo=None,
    ):
        Task.__init__(
            self,
            varinfo=varinfo,
            inputs=inputs,
            node_id=node_id,
            node_attrs=node_attrs,
            execinfo=execinfo,
        )
        if "recons_params" in inputs:
            raise KeyError("do not use 'recons_params' but use 'axis_params' instead")
        axis_params = inputs.get("axis_params", None)
        if isinstance(axis_params, dict):
            axis_params = AxisRP.from_dict(axis_params)
        elif not (axis_params is None or isinstance(axis_params, AxisRP)):
            raise TypeError(
                "'axis_params' is expected to be None or an instance of AxisRP or a dict"
            )

        SuperviseProcess.__init__(self, process_id=process_id)
        self._mode_calculation_fct = {}
        """dict with function pointer to call for making the mode calculation.
        Function should have only one 'scan' parameter as input"""

        self._axis_params = axis_params or AxisRP()
        """Axis reconstruction parameters to apply"""
        self._locked = False
        """Boolean used to lock reconstruction parameters edition"""
        self._recons_params_before_lock = None
        """Recons parameters register before locking the position"""

    def set_configuration(self, configuration):
        if "_rpSetting" in configuration:
            recons_params = AxisRP.from_dict(configuration["_rpSetting"])
        else:
            recons_params = AxisRP.from_dict(configuration)
        self.set_recons_params(recons_params=recons_params)

    def set_recons_params(self, recons_params):
        assert isinstance(recons_params, AxisRP)
        self._axis_params = recons_params

    def lock_position_value(self, lock=True):
        """
        lock the position currently computed or defined by the user.
        In this case we will lock the axis as defined 'fixed' with the current
        value

        :param lock: if true lock the currently existing position value
        """
        self._locked = lock
        if lock:
            self._recons_params_before_lock = self._axis_params.to_dict()
            if self._axis_params != AxisMode.manual:
                self._axis_params.mode = AxisMode.manual
        else:
            if self._recons_params_before_lock:
                self._axis_params.load_from_dict(
                    self._recons_params_before_lock
                )  # noqa

    def run(self):
        """
        Compute the position value then get ready to the next. And call

        .. note:: this simply call `compute`.
                  But this is needed for the AxisProcessThreaded class
        """
        scan = data_identifier_to_scan(self.inputs.data)
        if scan is None:
            self.outputs.data = None
            return

        if isinstance(scan, TomwerScanBase):
            scan = scan
        elif isinstance(scan, dict):
            scan = ScanFactory.create_scan_object_frm_dict(scan)
        else:
            raise ValueError(f"input type {scan} is not managed")

        _logger.info("start axis calculation for %s" % scan.path)
        self._axis_params.frame_width = scan.dim_1
        cor = error = None
        try:
            scan_res = self.compute(scan=scan)
        except Exception as e:
            scan_res = None
            error = e
        else:
            if isinstance(scan_res, TomwerScanBase):
                cor = scan_res.axis_params.relative_cor_value
            elif scan_res is None:
                if scan.axis_params.relative_cor_value is not None:
                    cor = scan.axis_params.relative_cor_value
            elif isinstance(scan_res, float):
                cor = scan_res
            else:
                assert isinstance(scan_res, dict)
                b_dict = scan_res
                if TomwerScanBase._DICT_AXIS_KEYS in scan_res:
                    b_dict = scan_res["axis_params"]
                cor = b_dict["POSITION_VALUE"]
        finally:
            if cor != "...":
                self._process_end(scan, cor=cor, error=error)

        if self.get_input_value("serialize_output_data", True):
            self.outputs.data = scan_res.to_dict()
        else:
            self.outputs.data = scan_res

    def _process_end(self, scan, cor, error=None):
        assert isinstance(scan, TomwerScanBase)
        try:
            extra = {
                logconfig.DOC_TITLE: self._scheme_title,
                logconfig.SCAN_ID: scan.path,
            }
            if error is not None:
                info = " ".join(
                    (
                        "fail to compute axis position for scan",
                        str(scan.path),
                        "reason is ",
                        str(error),
                    )
                )
                _logger.processFailed(info, extra=extra)
                ProcessManager().notify_dataset_state(
                    dataset=scan, process=self, state=DatasetState.FAILED, details=info
                )
            elif scan.axis_params.relative_cor_value is None:
                info = " ".join(
                    ("fail to compute axis position for scan", str(scan.path))
                )
                _logger.processFailed(info, extra=extra)
                ProcessManager().notify_dataset_state(
                    dataset=scan, process=self, state=DatasetState.FAILED, details=info
                )
            else:
                info = "axis calculation defined for {}: {} (using {})".format(
                    str(scan.path),
                    str(scan.axis_params.relative_cor_value),
                    scan.axis_params.mode.value,
                )
                _logger.processSucceed(info, extra=extra)
                ProcessManager().notify_dataset_state(
                    dataset=scan, process=self, state=DatasetState.SUCCEED, details=info
                )
        except Exception as e:
            _logger.error(e)

    @staticmethod
    def get_input_radios(scan) -> tuple:
        """Make sure we have valid projections to be used for axis calculation

        :param scan: scan to check
        :raise: NoAxisUrl if fails to found
        :return: tuple of AxisResource
        """
        if (
            scan.axis_params
            and scan.axis_params.axis_url_1
            and scan.axis_params.axis_url_1.url
        ):
            return scan.axis_params.axis_url_1, scan.axis_params.axis_url_2

        return scan.get_opposite_projections(mode=scan.axis_params.angle_mode)

    @staticmethod
    def get_inputs(scan):
        assert isinstance(scan, TomwerScanBase)
        radio_1, radio_2 = AxisTask.get_input_radios(scan=scan)
        if radio_1 and radio_2:
            mess = " ".join(
                ("input radios are", radio_1.url.path(), "and", radio_2.url.path())
            )
            _logger.info(mess)
            log_ = scan.axis_params.projection_type is ProjectionType.transmission

            # if necessary normalize data
            if radio_1.normalized_data is None:
                radio_1.normalize_data(scan, log_=log_)
            if radio_2.normalized_data is None:
                radio_2.normalize_data(scan, log_=log_)

            if scan.axis_params.paganin_preproc:
                data_1 = radio_1.normalized_data_paganin
                data_2 = radio_2.normalized_data_paganin
            else:
                data_1 = radio_1.normalized_data
                data_2 = radio_2.normalized_data

            if scan.axis_params.scale_img2_to_img1:
                data_2 = image.scale_img2_to_img1(img_1=data_1, img_2=data_2)
            return data_1, data_2
        else:
            _logger.info("fail to find any inputs")
            return None, None

    def compute(self, scan, wait=True):
        """
        Compute the position value for the scan

        :param scan:
        :param wait: used for threaded process. True if we want to end the
                          computation before releasing hand.
        :return: scan as a TomoBase
        """
        assert scan is not None
        if isinstance(scan, dict):
            _logger.warning("convert scan from a dict")
            _scan = ScanFactory.create_scan_object_frm_dict(scan)
        else:
            _scan = scan
        assert isinstance(_scan, TomwerScanBase)
        # if the scan has no tomo reconstruction parameters yet create them
        if _scan.axis_params is None:
            _scan.axis_params = AxisRP()

        # copy axis recons parameters. We skip the axis_url which are specific
        # to the scan
        _scan.axis_params.copy(
            self._axis_params, copy_axis_url=False, copy_flip_lr=False
        )
        assert scan.axis_params is not None
        return self._process_computation(scan=_scan)

    def scan_ready(self, scan):
        _logger.info(scan, "processed")

    def _process_computation(self, scan):
        """

        :param TomwerScanBase scan: scan for which we want to compute the axis
                              position.
        :return: scan as a TomoBase or a dict if serialize_output_data activated
        """
        _logger.info("compute center of rotation for %s" % scan.path)
        try:
            position = self.compute_axis_position(scan)
        except NotImplementedError as e:
            scan.axis_params.set_relative_value(None)
            raise e
        except ValueError as e:
            scan_name = scan.path or "undef scan"
            scan.axis_params.set_relative_value(None)
            raise Exception(
                f"Fail to compute axis position for {scan_name} reason is {e}"
            )
        else:
            scan.axis_params.set_relative_value(position)
            self._axis_params.frame_width = scan.dim_1
            self._axis_params.set_relative_value(position)
            scan_name = scan.path or "undef scan"
            r_cor_value = scan.axis_params.relative_cor_value
            mess = f"Compute axis position ({r_cor_value}) with {scan.axis_params.mode.value}"
            _logger.info(mess)
        return scan

    def setMode(self, mode, value):
        if mode is AxisMode.manual:
            self._axis_params.cor_position = value
        else:
            raise NotImplementedError("mode not implemented yet")

    def define_calculation_mode(self, mode: AxisMode, fct_pointer):
        """Register the function to call of the given mode

        :param mode: the mode to register
        :param fct_pointer: pointer to the function to call
        """
        self._mode_calculation_fct[mode] = fct_pointer

    def compute_axis_position(self, scan):
        """

        :param scan: scan for which we compute the center of rotation
        :return: position of the rotation axis. Use the `.AxisMode` defined
                 by the `.ReconsParams` of the `.AxisTask`
        """
        mode = self._axis_params.mode
        if mode is AxisMode.manual:
            # If mode is read or manual the position_value is not computed and
            # we will keep the actual one (should have been defined previously)
            res = self._axis_params.relative_cor_value
        elif mode is AxisMode.read:
            res = read_scan_x_rotation_axis_pixel_position(scan=scan)
        else:
            has_darks = scan.reduced_darks is not None and len(scan.reduced_darks) > 0
            has_flats = scan.reduced_flats is not None and len(scan.reduced_flats) > 0
            do_flatfield = has_darks and has_flats
            res = estimate_cor(
                method=mode.value,
                dataset_info=adapt_tomwer_scan_to_nabu(
                    scan=scan, do_flatfield=do_flatfield
                ),
                do_flatfield=do_flatfield,
                cor_options=scan.axis_params.get_nabu_cor_options_as_dict(),
            )
            # convert back to relative
            res = _absolute_pos_to_relative_with_warning(
                absolute_pos=res, det_width=scan.dim_1
            )

        return res

    @docstring(Task.program_name)
    @staticmethod
    def program_name():
        return "tomwer_axis"

    @docstring(Task.program_version)
    @staticmethod
    def program_version():
        return tomwer.version.version

    @docstring(Task.definition)
    @staticmethod
    def definition():
        return "Compute center of rotation"



# coding: utf-8
from __future__ import annotations

import logging
import os
import pickle

from orangecanvas.scheme.readwrite import literal_dumps
from processview.core.manager import DatasetState, ProcessManager
from silx.gui import qt

from orangecontrib.tomwer.widgets.reconstruction.SinoNormOW import (
    SinoNormOW as _SinoNormOW,
)
from tomwer.core.utils.scanutils import MockNXtomo
from tomwer.tests.conftest import qtapp  # noqa F401

logger = logging.getLogger(__name__)


class SinoNormOW(_SinoNormOW):
    def __init__(self, parent=None):
        self._scans_finished = []
        super().__init__(parent)

    def processing_finished(self, scan):
        self._scans_finished.append(scan)

    def wait_processing(self, wait_time):
        self._window._processing_stack._computationThread.wait(wait_time)

    @property
    def scans_finished(self):
        return self._scans_finished

    def compute(self):
        self._window._processCurrentScan()

    def setROI(self, start_x, end_x, start_y, end_y):
        self._window.setROI(start_x=start_x, end_x=end_x, start_y=start_y, end_y=end_y)

    def close(self):
        self._scans_finished = {}
        super().close()


FRAME_DIM = 100


def create_scan(folder_name):
    return MockNXtomo(
        scan_path=folder_name,
        n_ini_proj=20,
        n_proj=20,
        n_alignement_proj=2,
        create_final_flat=False,
        create_ini_dark=True,
        create_ini_flat=True,
        n_refs=1,
        dim=FRAME_DIM,
    ).scan


def test_SinoNormOW(
    qtapp,  # noqa F811
    tmp_path,
):
    source_dir = tmp_path / "source"
    source_dir.mkdir()

    # create scans
    scan_1 = create_scan(os.path.join(source_dir, "scan_1"))
    scan_2 = create_scan(os.path.join(source_dir, "scan_2"))
    scan_3 = create_scan(os.path.join(source_dir, "scan_3"))
    process_manager = ProcessManager()

    widget = SinoNormOW()
    widget.show()

    # test serialization
    pickle.dumps(widget.getConfiguration())

    # test literal dumps
    widget._updateSettings()
    literal_dumps(widget._ewoks_default_inputs)

    # test behavior when the widget is unlocked
    """Test result when used with some interaction"""
    widget.setLocked(False)

    def process_scalar_manually():
        widget.setCurrentMethod("division")
        widget.setCurrentSource("manual ROI")

        qt.QApplication.processEvents()
        widget.setROI(start_x=0, end_x=10, start_y=0, end_y=10)
        qt.QApplication.processEvents()
        widget.compute()
        widget.wait_processing(5000)
        qt.QApplication.processEvents()

    widget.process(scan_1)
    process_scalar_manually()
    assert (
        process_manager.get_dataset_state(
            dataset_id=scan_1.get_identifier(),
            process=widget,
        )
        == DatasetState.WAIT_USER_VALIDATION
    )

    widget.process(scan_2)
    process_scalar_manually()
    assert len(widget.scans_finished) == 0
    assert (
        process_manager.get_dataset_state(
            dataset_id=scan_1.get_identifier(),
            process=widget,
        )
        == DatasetState.SKIPPED
    )
    assert (
        process_manager.get_dataset_state(
            dataset_id=scan_2.get_identifier(),
            process=widget,
        )
        == DatasetState.WAIT_USER_VALIDATION
    )

    widget.process(scan_3)
    process_scalar_manually()
    widget.validateCurrentScan()
    assert (
        process_manager.get_dataset_state(
            dataset_id=scan_3.get_identifier(),
            process=widget,
        )
        == DatasetState.SUCCEED
    )

    # test behavior when the widget is locked

    widget.setLocked(True)
    for scan in (scan_1, scan_2, scan_3):
        widget.process(scan)
        widget.wait_processing(5000)
        qt.QApplication.processEvents()

    for scan in (scan_1, scan_2, scan_3):
        # test status is SUCCEED
        assert (
            process_manager.get_dataset_state(
                dataset_id=scan.get_identifier(),
                process=widget,
            )
            == DatasetState.SUCCEED
        )



# coding: utf-8
"""module containing FutureSupervisorTask."""

from typing import Iterable

from ewokscore.task import Task as EwoksTask
from tomwer.core.futureobject import FutureTomwerObject
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.process.drac.processeddataset import DracReconstructedVolumeDataset
from tomwer.core.volume.volumebase import TomwerVolumeBase
from tomwer.core.volume.volumefactory import VolumeFactory


class FutureSupervisorTask(
    EwoksTask,
    input_names=("future_tomo_obj",),
    output_names=("data", "volume", "data_portal_processed_datasets"),
):
    """
    Task used to wait for a 'FutureTomwerObject' and convert it to original instance of:

    * TomwerScanBase (data): if the FutureTomwerObject is based on a scan instance
    * TomwerVolumeBase (volume): if the FutureTomwerObject is based on a volume instance
    * tuple of IcatReconstructedVolumeDataset (data_portal_processed_datasets): if the FutureTomwerObject is based on a volume instance
    """

    def run(self):
        future_tomo_obj = self.inputs.future_tomo_obj
        if not isinstance(future_tomo_obj, FutureTomwerObject):
            raise TypeError(
                f"future_data is expected to be an instance of {FutureTomwerObject}. Got {type(future_tomo_obj)}"
            )
        future_tomo_obj.results()

        tomo_obj = future_tomo_obj.tomo_obj

        if isinstance(tomo_obj, TomwerScanBase):
            self.outputs.data = tomo_obj

            # the volume reconstruction return an instance of 'FutureTomwerObj' with a scan.
            # so to make drac publication compatible with it we also need to send the 'data_portal_processed_datasets'
            def build_drac_dataset(vol_id):
                volume = VolumeFactory.create_tomo_object_from_identifier(vol_id)
                return DracReconstructedVolumeDataset(
                    tomo_obj=volume,
                    source_scan=tomo_obj,
                )

            assert isinstance(tomo_obj.latest_vol_reconstructions, Iterable)
            self.outputs.data_portal_processed_datasets = tuple(
                [
                    build_drac_dataset(vol_id=vol_id)
                    for vol_id in tomo_obj.latest_vol_reconstructions
                ]
            )
            if len(tomo_obj.latest_vol_reconstructions) > 0:
                # at the moment we expect at most one volume from a scan.
                # If it changes we will need to replace `volume` by `volumes`
                self.outputs.volume = VolumeFactory.create_tomo_object_from_identifier(
                    tomo_obj.latest_vol_reconstructions[0]
                )
            else:
                self.outputs.volume = None
        elif isinstance(tomo_obj, TomwerVolumeBase):
            self.outputs.volume = tomo_obj
            self.outputs.data_portal_processed_datasets = ()
            self.outputs.data = None



"""
(Ewoks)Orange Widgets dedicated for tomography
"""

from ewoksorange.pkg_meta import get_distribution


ICON = "../widgets/icons/tomwer.png"

BACKGROUND = "#C0CCFF"


# Entry point for main Orange categories/widgets discovery
def widget_discovery(discovery):
    dist = get_distribution("tomwer")
    pkgs = [
        "orangecontrib.tomwer.widgets.cluster",
        "orangecontrib.tomwer.widgets.control",
        "orangecontrib.tomwer.widgets.deprecated",
        "orangecontrib.tomwer.widgets.debugtools",
        "orangecontrib.tomwer.widgets.edit",
        "orangecontrib.tomwer.widgets.dataportal",
        "orangecontrib.tomwer.widgets.reconstruction",
        # "orangecontrib.tomwer.widgets.stitching",
        "orangecontrib.tomwer.widgets.visualization",
        "orangecontrib.tomwer.widgets.other",
    ]
    for pkg in pkgs:
        discovery.process_category_package(pkg, distribution=dist)


WIDGET_HELP_PATH = (
    # Used for development.
    # You still need to build help pages using
    # make htmlhelp
    # inside doc folder
    # (
    #     "/home/payno/Documents/dev/tomography/tomwer/build/html/canvas/widgets/widgets.html",
    #     None,
    # ),
    ("https://tomwer.readthedocs.io/en/latest/canvas/widgets/widgets.html", None),
)



# coding: utf-8
from __future__ import annotations


from silx.gui import qt
from tomwer.gui.cluster.slurm import SlurmSettingsWidget


class SlurmSettingsDialog(qt.QDialog):
    sigConfigChanged = qt.Signal()
    """Signal emit when the SlurmSetting changed"""

    def __init__(self, parent=None):
        super().__init__(parent=parent)

        self.setLayout(qt.QVBoxLayout())
        self._mainWidget = SlurmSettingsWidget(parent=self)
        self.layout().addWidget(self._mainWidget)

        # buttons for validation
        self._buttons = qt.QDialogButtonBox(self)
        types = qt.QDialogButtonBox.Close
        self._buttons.setStandardButtons(types)
        self.layout().addWidget(self._buttons)

        self._buttons.button(qt.QDialogButtonBox.Close).clicked.connect(self.close)

        # connect signal /slot
        self._mainWidget.sigConfigChanged.connect(self._configChanged)

    def _configChanged(self, *args, **kwargs):
        self.sigConfigChanged.emit()

    def isSlurmActive(self):
        return self._mainWidget.isSlurmActive()

    def getConfiguration(self) -> dict:
        return self._mainWidget.getConfiguration()

    def setConfiguration(self, config: dict) -> None:
        self._mainWidget.setConfiguration(config=config)



"""contains dialogs to select a Scan (aka Data)"""

from __future__ import annotations

import logging
import os

from silx.gui import qt


_logger = logging.getLogger()


class QDataDialog(qt.QFileDialog):
    """dialog to select a scans (aka data)"""

    def __init__(self, parent, multiSelection=False):
        qt.QFileDialog.__init__(self, parent)
        self.setNameFilters(
            [
                "HDF5 files (*.h5 *.hdf5 *.nx *.nxs *.nexus)",
                "Nexus files (*.nx *.nxs *.nexus)",
                "Any files (*)",
            ]
        )
        self._selected_files = []
        self.setFileMode(qt.QFileDialog.ExistingFiles)

        # self.QDialogButtonBox
        self.multiSelection = multiSelection
        # check if 'TOMWER_DEFAULT_INPUT_DIR' has been set
        if os.environ.get("TOMWER_DEFAULT_INPUT_DIR", None) and os.path.exists(
            os.environ["TOMWER_DEFAULT_INPUT_DIR"]
        ):
            self.setDirectory(os.environ["TOMWER_DEFAULT_INPUT_DIR"])
        elif self.directory() != os.getcwd() or str(self.directory()).startswith(
            "/data"
        ):
            # if the directory as already been set by the user. Avoid redefining it
            pass
        elif os.path.isdir("/data"):
            self.setDirectory("/data")

        btns = self.findChildren(qt.QPushButton)
        if len(btns) == 0:
            _logger.error(
                "Cannot retrieve open button. switch to none " "native QFileDialog"
            )
            self.setOption(qt.QFileDialog.DontUseNativeDialog)
            btns = self.findChildren(qt.QPushButton)

        if self.multiSelection is True:
            # to make it possible to select multiple directories:
            self.file_view = self.findChild(qt.QListView, "listView")
            if self.file_view:
                self.file_view.setSelectionMode(qt.QAbstractItemView.MultiSelection)
                self.file_view.setSelectionMode(qt.QAbstractItemView.ExtendedSelection)

            self.f_tree_view = self.findChild(qt.QTreeView)
            if self.f_tree_view:
                self.f_tree_view.setSelectionMode(qt.QAbstractItemView.MultiSelection)
                self.f_tree_view.setSelectionMode(
                    qt.QAbstractItemView.ExtendedSelection
                )

            if len(btns) > 0:
                self.openBtn = [x for x in btns if "open" in str(x.text()).lower()][0]
                self.openBtn.clicked.disconnect()
                self.openBtn.hide()
                parent = self.openBtn.parent()
                self.openBtn = qt.QPushButton("Select", parent=parent)
                self.openBtn.clicked.connect(self.openClicked)
                parent.layout().insertWidget(0, self.openBtn)

    def openClicked(self):
        inds = self.f_tree_view.selectionModel().selectedIndexes()
        for i in inds:
            if i.column() == 0:
                self._selected_files.append(
                    os.path.join(str(self.directory().absolutePath()), str(i.data()))
                )
        self.accept()
        self.done(1)

    def files_selected(self):
        for file_ in self.selectedFiles():
            self._selected_files.append(file_)
        return self._selected_files



import os

from tomwer.core.scan.blissscan import BlissScan
from tomwer.core.scan.edfscan import EDFTomoScan
from tomwer.core.scan.nxtomoscan import NXtomoScan
from tomwer.core.utils.scanutils import format_output_location


def test_format_output_location(tmp_path):
    """
    test different use cases of 'format_output_location'
    """
    bliss_raw_dir = tmp_path / "raw"
    bliss_raw_dir.mkdir()
    bliss_sample_dir = bliss_raw_dir / "sample"
    bliss_sample_dir.mkdir()
    bliss_master_file = bliss_sample_dir / "dataset.h5"

    bliss_proposal_file = bliss_raw_dir / "ihsample.h5"
    bliss_proposal_file = os.path.abspath(bliss_proposal_file)

    bliss_scan = BlissScan(
        master_file=bliss_master_file, entry="1.1", proposal_file=bliss_proposal_file
    )
    expected_path = os.path.join(tmp_path, "reduced", "sample")
    assert (
        format_output_location(
            location="{scan_parent_dir_basename}/../reduced/{scan_dir_name}",
            scan=bliss_scan,
        )
        == expected_path
    )

    edf_scan = EDFTomoScan("/test/my/folder/")
    assert (
        format_output_location(
            location="{scan_parent_dir_basename}/output", scan=edf_scan
        )
        == "/test/my/output"
    )

    hdf5_scan = NXtomoScan("/ddsad/my/file.hdf5", entry="nxtomos/entry0000")
    assert (
        format_output_location(location="{scan_basename}/output.nx", scan=hdf5_scan)
        == "/ddsad/my/output.nx"
    )
    assert format_output_location(
        location="{scan_file_name}/{scan_entry}.nx", scan=hdf5_scan
    ) == os.path.abspath("file/nxtomos_entry0000.nx")



# coding: utf-8

ALPHA_CHAR = (b"\xce\xb1").decode("utf-8")

BETA_CHAR = (b"\xce\xb2").decode("utf-8")

GAMMA_CHAR = (b"\xce\xb3").decode("utf-8")

DELTA_CHAR = (b"\xce\xb4").decode("utf-8")

PSI_CHAR = (b"\xcf\x88").decode("utf-8")

THETA_CHAR = (b"\xce\xb8").decode("utf-8")

MU_CHAR = (b"\xce\xbc").decode("utf-8")

DEGREE_CHAR = (b"\xc2\xb0").decode("utf-8")



from __future__ import annotations

from pydantic import field_validator
from ewokscore.model import BaseInputModel
from tomwer.core.volume.volumebase import TomwerVolumeBase
from tomwer.core.volume.volumefactory import VolumeFactory


class VolumeViewerModel(BaseInputModel):
    volume: TomwerVolumeBase | str
    load_volume: bool

    @field_validator("volume", mode="before")
    @classmethod
    def cast_volume(cls, value) -> TomwerVolumeBase:
        if isinstance(value, str):
            value = VolumeFactory.create_tomo_object_from_identifier(value)
        return value



# coding: utf-8
from __future__ import annotations


import os
import shutil
import tempfile
import unittest

import pytest
from nxtomo.nxobject.nxdetector import ImageKey

from tomwer.core.process.edit.imagekeyeditor import (
    ImageKeyEditorTask,
    ImageKeyUpgraderTask,
)
from tomwer.core.scan.nxtomoscan import NXtomoScan
from tomwer.core.utils.scanutils import MockNXtomo


class TestImageKeyEditor(unittest.TestCase):
    def setUp(self):
        self.tempdir = tempfile.mkdtemp()
        dim = 10
        mock = MockNXtomo(
            scan_path=os.path.join(self.tempdir, "scan1"),
            n_proj=10,
            n_ini_proj=10,
            scan_range=180,
            dim=dim,
        )
        self.scan = mock.scan

    def tearDown(self):
        shutil.rmtree(self.tempdir)

    def test(self):
        process = ImageKeyEditorTask(
            inputs={
                "data": self.scan,
                "serialize_output_data": False,
                "configuration": {
                    "modifications": {
                        0: ImageKey.INVALID.value,
                        3: ImageKey.ALIGNMENT.value,
                    },
                },
            }
        )
        process.definition()
        process.program_version()
        process.program_name()
        process.run()


def test_ImageKeyUpgraderTask(tmp_path):
    """
    test ImageKeyUpgraderTask task
    """
    test_dir = tmp_path / "test_image_key_upgrader"
    os.makedirs(test_dir)
    scan = MockNXtomo(
        scan_path=test_dir,
        n_proj=20,
        n_ini_proj=20,
        create_ini_dark=False,
    ).scan

    operations = {
        ImageKey.PROJECTION: ImageKey.DARK_FIELD,
    }

    with pytest.raises(TypeError):
        ImageKeyUpgraderTask(
            inputs={
                "data": None,
                "operations": operations,
            },
        ).run()

    with pytest.raises(TypeError):
        ImageKeyUpgraderTask(
            inputs={
                "data": scan,
                "operations": None,
            },
        ).run()

    task = ImageKeyUpgraderTask(
        inputs={
            "data": scan,
            "operations": operations,
        },
    )
    task.run()

    scan = NXtomoScan(scan.master_file, scan.entry)
    assert len(scan.projections) == 0
    assert len(scan.darks) == 20



from __future__ import annotations

import os
import shutil
import tempfile
import pint

from silx.gui import qt
from silx.gui.utils.testutils import TestCaseQt

from tomwer.gui.stitching.tests.utils import create_scans_z_series
from tomwer.gui.stitching.StitchingWindow import ZStitchingWindow


_ureg = pint.get_application_registry()


class TestZStichingWindow(TestCaseQt):
    """
    Test high level z stitching definition
    """

    def setUp(self):
        super().setUp()
        self._tmp_path = tempfile.mkdtemp()
        self.scans = create_scans_z_series(
            os.path.join(self._tmp_path, "case1"),
            z_positions_m=(0.200, 0.205, 0.210),
            sample_pixel_size=0.001,
            raw_frame_width=100,
        )

        self._widget = ZStitchingWindow()

    def tearDown(self):
        self._widget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self._widget.close()
        self._widget = None
        while self.qapp.hasPendingEvents():
            self.qapp.processEvents()
        shutil.rmtree(self._tmp_path)

    @staticmethod
    def checkNbReceivers(scan, nb_receivers):
        return (
            scan.stitching_metadata.receivers(scan.stitching_metadata.sigChanged)
            == nb_receivers
        )

    def test(self):
        self._widget.show()
        # fill the widget with scans

        for scan in self.scans:
            self._widget.addTomoObj(scan)
            # note: add the sigChanged to the z ordered list 'update z' slot



from __future__ import annotations

import logging
from collections import OrderedDict
from typing import Iterable

import numpy
from silx.gui import qt
from silx.gui.plot import PlotWidget

logger = logging.getLogger(__name__)


class QSliceSelectorDialog(qt.QDialog):
    """
    The dialog used to select some slice indexes from a radio

    :param QWidget parent: parent widget
    :param n_required_slice: number of required slice we expect the user to select
    """

    def __init__(self, parent, n_required_slice: int | None = None):
        qt.QDialog.__init__(self, parent=parent)
        self.setLayout(qt.QVBoxLayout())
        self.setWindowTitle("select slices on radio")

        self.mainWidget = QLineSelector(parent=self, n_required_slice=n_required_slice)
        self.layout().addWidget(self.mainWidget)

        types = qt.QDialogButtonBox.Ok | qt.QDialogButtonBox.Cancel
        self._buttons = qt.QDialogButtonBox(parent=self)
        self._buttons.setStandardButtons(types)
        self.layout().addWidget(self._buttons)

        # connect signal / slots
        self._buttons.button(qt.QDialogButtonBox.Ok).clicked.connect(self.accept)
        self._buttons.button(qt.QDialogButtonBox.Cancel).clicked.connect(self.reject)

        # expose API
        self.setData = self.mainWidget.setData

    def setSelection(self, selection: Iterable):
        """

        :param rows: define the selection
        """
        if self.nRequiredSlice() is not None and len(selection) > self.nRequiredSlice():
            raise ValueError(
                f"you provide a selection of {len(selection)} elements when the user should select {self._n_required_slice}"
            )
        if type(selection) is str:
            selection = selection.replace("(", "")
            selection = selection.replace(")", "")
            selection = selection.replace(" ", "")
            selection = selection.replace(",", ";")
            try:
                selection = [int(item) for item in selection.split(";")]
            except Exception as e:
                logger.error(f"Fail to set selection. Error is {e}")
        self.mainWidget.setSelection(selection)

    def getSelection(self) -> tuple:
        """

        :return: the selection of slices to use
        """
        return self.mainWidget.getSelection()

    def exec_(self):
        if not self.mainWidget._has_data:
            mess = "no data set, can't use the selection tool"
            logger.warning(mess)
            qt.QMessageBox.warning(self, "Selection tool not available", mess)
            self.reject()
        else:
            return qt.QDialog.exec(self)

    def nRequiredSlice(self) -> int | None:
        return self.mainWidget.nRequiredSlice()


class QLineSelector(qt.QWidget):
    """Widget to select a set of slices from a plot"""

    def __init__(self, parent=None, n_required_slice=None):
        self._n_required_slice = n_required_slice
        qt.QWidget.__init__(self, parent)
        self._plot = PlotWidget()
        # invert y axis
        self._plot.setYAxisInverted(True)
        self.__selection = OrderedDict()
        # dict of markers from the user selection. Keys are item legend, value
        # are items
        self._has_data = False
        self.setLayout(qt.QVBoxLayout())
        self.layout().addWidget(self._plot)
        # connect signal / slot
        self._plot.sigPlotSignal.connect(self._plotDrawEvent)

    def nRequiredSlice(self) -> int | None:
        return self._n_required_slice

    def setData(self, data: numpy.ndarray):
        """
        Define the data from which we can select slices

        :param data: data to plot
        """
        self._has_data = True
        self._plot.addImage(data)
        assert self._plot.getActiveImage(just_legend=True) is not None

    def getSelection(self) -> tuple:
        """

        :return: the selection of slices to use
        """
        res = []
        for _, marker in self.__selection.items():
            res.append(int(marker.getPoints()[0][1]))
        return tuple(sorted(res))

    def _clear(self):
        if len(self.__selection) > 0:
            self.removeSlice(list(self.__selection.keys())[0])
            self._clear()

    def nSelected(self) -> int:
        """Return the number of slice selected"""
        return len(self.__selection)

    def setSelection(self, rows: Iterable) -> None:
        """

        :param rows: define the selection
        """
        assert type(rows) is not str
        self._clear()
        if rows is not None:
            for row in rows:
                self.addSlice(row)

    def addSlice(self, row: float | int) -> None:
        """
        Add the requested slice to the selection

        :param row:
        """
        row_n = round(float(row))

        if self._plot.getActiveImage(just_legend=True) is not None:
            data_bounds = self._plot.getActiveImage(just_legend=False).getBounds()
            if not (data_bounds[0] <= row_n < data_bounds[1]):
                logger.warning("requested slice out of the data, ignored")
                return

        # remove the n first selected point if needed
        n_selected = len(self.getSelection())
        n_required_slice = self.nRequiredSlice()
        if (n_required_slice is not None) and (n_selected >= n_required_slice):
            n_slice_to_rm = n_required_slice - n_selected + 1
            # +1 because we are about to add a slice
            markers = tuple(self.__selection.values())
            idx_to_rm = [
                int(marker.getPoints()[0][1]) for marker in markers[0:n_slice_to_rm]
            ]
            for slice_idx in idx_to_rm:
                self.removeSlice(slice_idx)

        inf = 10000
        legend = self._getLegend(row_n=row_n)
        self._plot.addShape(
            xdata=numpy.array((-inf, -inf, inf, inf)),
            ydata=numpy.array((row_n, row_n + 1, row_n + 1, row_n)),
            shape="polygon",
            linewidth=1,
            color="pink",
            linestyle="-",
            fill=True,
            legend=legend,
        )
        self.__selection[legend] = self._plot._getItem(kind="item", legend=legend)

    def _getLegend(self, row_n) -> str:
        return str(round(float(row_n)))

    def removeSlice(self, row: float | int) -> None:
        """
        remove the requested slice from the selection

        :param row: row containing the slice to be removed
        """
        legend = self._getLegend(row_n=row)
        if legend in self.__selection:
            self._plot.remove(legend=legend, kind="item")
            del self.__selection[legend]

    def _plotDrawEvent(self, event):
        if "event" in event and event["event"] == "mouseClicked":
            row = event["y"] - 0.5

            if qt.QApplication.keyboardModifiers() & qt.Qt.ControlModifier:
                self.removeSlice(row=row)
            else:
                self.addSlice(row=row)



import pytest
from orangewidget import gui, widget
from orangewidget.widget import Input

from silx.gui import qt

import tomwer.core.process.visualization.volumeviewer
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.volume.volumebase import TomwerVolumeBase
from ._volume_viewer import VolumeViewer


@pytest.mark.skip("Fail on CI")
class VolumeViewerOW(widget.OWBaseWidget, openclass=True):
    """a viewer to display the last volume reconstructed using silx plot3d
    viewer.

    :param parent: the parent widget
    """

    name = "volume viewer"
    id = "orange.widgets.tomwer.deprecated.VolumeViewerOW.VolumeViewerOW"
    description = "display the last volume reconstructed"
    icon = "icons/deprecated_volume_viewer.svg"
    priority = 80
    keywords = ["tomography", "file", "tomwer", "acquisition", "validation"]

    ewokstaskclass = (
        tomwer.core.process.visualization.volumeviewer._VolumeViewerPlaceHolder
    )

    want_main_area = True
    want_control_area = False
    resizing_enabled = True

    class Inputs:
        data = Input(name="data", type=TomwerScanBase)
        volume = Input(name="volume", type=TomwerVolumeBase, multiple=True)

    def __init__(self, parent=None):
        super().__init__(parent)
        self._layout = gui.vBox(self.mainArea, self.name).layout()
        self.viewer = VolumeViewer(parent=self)
        self._layout.addWidget(self.viewer)

    @Inputs.data
    def addScan(self, scan, *args, **kwargs):
        if scan is None:
            return
        self.viewer.setScan(scan)

    @Inputs.volume
    def _volumeReceived(self, volume, *args, **kwargs):
        self.addVolume(volume)

    def addVolume(self, volume):
        if volume is None:
            return
        self.viewer.setVolume(volume)

    def sizeHint(self):
        return qt.QSize(400, 500)



import logging
import os
import pickle

import numpy
from orangecanvas.scheme.readwrite import literal_dumps
from processview.core.manager import DatasetState, ProcessManager
from silx.io.url import DataUrl
from silx.gui import qt

from orangecontrib.tomwer.widgets.reconstruction.SADeltaBetaOW import (
    SADeltaBetaOW as _SADeltaBetaOW,
)
from tomwer.core.process.reconstruction.scores import ComputedScore
from tomwer.core.utils.scanutils import MockNXtomo
from tomwer.tests.conftest import qtapp  # noqa F401

logger = logging.getLogger(__name__)


class SADeltaBetaOW(_SADeltaBetaOW):
    """class saving scan finished to keep a trace on those."""

    def __init__(self, parent=None):
        self._scans_finished = []
        super().__init__(parent)

    def processing_finished(self, scan):
        self._scans_finished.append(scan)

    @property
    def scans_finished(self):
        return self._scans_finished

    def close(self):
        self._scans_finished = {}
        super().close()


FRAME_DIM = 100


def create_scan(output_dir):
    return MockNXtomo(
        scan_path=output_dir,
        n_ini_proj=20,
        n_proj=20,
        n_alignement_proj=2,
        create_final_flat=False,
        create_ini_dark=True,
        create_ini_flat=True,
        n_refs=1,
        dim=FRAME_DIM,
    ).scan


def patch_score(*args, **kwargs):
    """Function to save some result"""
    return DataUrl(
        file_path="/no_existing/path.hdf5",
        data_path="/no_existing_data_path",
        scheme="silx",
    ), ComputedScore(
        tv=numpy.random.random(),
        std=numpy.random.random(),
    )


def test_SADeltaBetaOW(
    qtapp,  # noqa F811
    tmp_path,
):
    source_dir = tmp_path / "source"
    source_dir.mkdir()

    # create scans
    scan_1 = create_scan(os.path.join(source_dir, "scan_1"))
    scan_2 = create_scan(os.path.join(source_dir, "scan_2"))
    scan_3 = create_scan(os.path.join(source_dir, "scan_3"))

    process_manager = ProcessManager()

    widget = SADeltaBetaOW()
    widget.show()

    widget._widget._processing_stack.patch_processing(patch_score)

    # make sure the configuration is serializable
    pickle.dumps(widget.getConfiguration())

    # make sure this is compatible with 'literal dumps'
    literal_dumps(widget.getConfiguration())

    # test behavior when 'autofocus' is lock. (in this case we expect to take the dataset with the higher score) and continue
    widget.lockAutofocus(False)

    def manual_processing():
        widget.compute()
        qt.QApplication.processEvents()
        widget.wait_processing(5000)
        qt.QApplication.processEvents()

    widget.process(scan_1)
    manual_processing()
    assert (
        process_manager.get_dataset_state(
            dataset_id=scan_1.get_identifier(),
            process=widget,
        )
        == DatasetState.WAIT_USER_VALIDATION
    )

    widget.process(scan_2)
    manual_processing()
    assert len(widget.scans_finished) == 0
    assert (
        process_manager.get_dataset_state(
            dataset_id=scan_1.get_identifier(),
            process=widget,
        )
        == DatasetState.SKIPPED
    )

    assert (
        process_manager.get_dataset_state(
            dataset_id=scan_2.get_identifier(),
            process=widget,
        )
        == DatasetState.WAIT_USER_VALIDATION
    )

    widget.process(scan_3)
    manual_processing()
    widget.validateCurrentScan()
    assert (
        process_manager.get_dataset_state(
            dataset_id=scan_3.get_identifier(),
            process=widget,
        )
        == DatasetState.SUCCEED
    )

    # ensure a cor has been registered
    assert scan_3.sa_delta_beta_params.selected_delta_beta_value is not None



from __future__ import annotations

import sys
from silx.gui import qt
import functools

import signal

from ..control._DataListenerBaseClass import DataListenerBaseClass
from tomwer.core.process.control.datalistener import DataListener
from tomwer.synctools.bliss_sync.datalistener import (
    DataListenerQThread,
    MockDataListenerQThread,
)
from tomwer.gui.dialog._PortOccupyDialog import PortOccupyDialog
from tomwer.gui.utils.qt_utils import block_signals


class DataListenerOW(DataListenerBaseClass):
    """
    This widget is used to listen to a server notifying the widget when an
    acquisition is finished.
    Then the bliss file will be converted to .nx file, NXtomo compliant.
    """

    name = "Old scan listener"
    id = "orangecontrib.widgets.tomwer.deprecated.DataListenerOW.DataListenerOW"
    description = (
        "The widget will receive information from bliss acquisition "
        "and wait for acquisition to be finished. Once finished it "
        "will call nxtomomill to convert from bliss .hdf5 to "
        "NXtomo compliant .nx file"
    )
    icon = "icons/datalistener.svg"
    priority = 10
    keywords = [
        "tomography",
        "file",
        "tomwer",
        "listener",
        "datalistener",
        "hdf5",
        "NXtomo",
    ]

    def __init__(self, host_discovery=None, parent=None):
        super().__init__(host_discovery=host_discovery, parent=parent, uses_rpc=True)

        # manage server stop when delete directly the widget or stop by Ctr+C
        signal.signal(signal.SIGINT, self.handleSigTerm)
        onDestroy = functools.partial(self._stopServerBeforeClosing)
        self.destroyed.connect(onDestroy)

    def delete_listening_thread(self):
        if self._listening_thread is not None:
            self._listening_thread.sigAcquisitionStarted.disconnect(
                self._widget._acquisitionStarted
            )
            self._listening_thread.sigAcquisitionEnded.disconnect(
                self._widget._acquisitionEnded
            )
            self._listening_thread.sigScanAdded.disconnect(
                self._widget._acquisitionUpdated
            )
            self._listening_thread.sigServerStop.disconnect(self._widget._serverStopped)
        DataListener.delete_listening_thread(self)

    def create_listening_thread(self):
        if self._mock is True:
            thread = MockDataListenerQThread(
                host=self.getHost(),
                port=self.getPort(),
                acquisitions=None,
                mock_acquisitions=self._mock_acquisitions,
            )
        else:
            thread = DataListenerQThread(
                host=self.getHost(), port=self.getPort(), acquisitions=None
            )
        # connect thread
        thread.sigAcquisitionStarted.connect(
            self._widget._acquisitionStarted, qt.Qt.DirectConnection
        )
        thread.sigAcquisitionEnded.connect(
            self._widget._acquisitionEnded, qt.Qt.DirectConnection
        )
        thread.sigScanAdded.connect(
            self._widget._acquisitionUpdated, qt.Qt.DirectConnection
        )
        thread.sigServerStop.connect(
            self._widget._serverStopped, qt.Qt.DirectConnection
        )
        return thread

    def _stopServerBeforeClosing(self):
        self.activate(False)

    def handleSigTerm(self, signo, *args, **kwargs):
        if signo == signal.SIGINT:
            self._stopServerBeforeClosing()
            sys.exit()

    def activate(self, activate=True, is_using_rpc=True):
        """Overwrite the activate to notify the user in case the port is occupy"""
        if activate and not self.is_port_available():
            with block_signals(self._widget):
                self._widget.activate(activate=False)
            dialog = PortOccupyDialog(parent=self, port=self.port, host=self.host)
            dialog.setModal(False)
            if dialog.exec() == qt.QDialog.Accepted:
                if dialog.retry_connection:
                    return self.activate(activate=True, is_using_rpc=is_using_rpc)
            else:
                return

        super().activate(activate=activate, is_using_rpc=is_using_rpc)



from __future__ import annotations

import logging
import shutil

from silx.gui import qt
from silx.gui.dialog.DataFileDialog import DataFileDialog
from silx.io.url import DataUrl

from tomwer.core import settings
from tomwer.core.utils.lbsram import is_low_on_memory
from tomwer.core.process.reconstruction.darkref.darkrefscopy import DarkRefsCopy
from tomwer.gui.reconstruction.darkref.darkrefwidget import DarkRefWidget
from tomwer.io.utils import get_default_directory

_logger = logging.getLogger(__name__)


class DarkRefAndCopyWidget(DarkRefWidget):
    """
    Widget associated to the DarkRefCopy process
    """

    sigModeAutoChanged = qt.Signal()
    """Signal emitted when the mode auto change"""
    sigCopyActivationChanged = qt.Signal()
    """Signal emitted when the copy is activated or deactivated"""
    sigClearCache = qt.Signal()

    def __init__(self, save_dir: str, parent=None, reconsparams=None, process_id=None):
        DarkRefWidget.__init__(
            self, parent=parent, reconsparams=reconsparams, process_id=process_id
        )
        self._mode_auto = True

        self._refCopyWidget = RefCopyWidget(
            parent=self, save_dir=save_dir, refCopy=self
        )
        iCopy = self.mainWidget.addTab(self._refCopyWidget, "copy")
        tooltip = (
            "When copy is activated it will record refHST and dark "
            "files. \n Then when an acquisition without dark or refHST "
            "go through the widget it will copy those dark and refHST "
            "in the acquisition."
        )
        self.mainWidget.setTabToolTip(iCopy, tooltip)
        self.layout().setContentsMargins(0, 0, 0, 0)

        # connect signal / slot
        self._refCopyWidget.sigModeAutoChanged.connect(self._triggerModeAuto)
        self._refCopyWidget.sigCopyActivationChanged.connect(
            self._triggerCopyActivation
        )
        self._refCopyWidget.sigClearCache.connect(self.sigClearCache)

    def setRefSetBy(self, scan_id: str):
        self._refCopyWidget._statusBar.showMessage(f"ref set from {scan_id}")

    def set_mode_auto(self, auto):
        self._mode_auto = auto

    def setRefsFromScan(self, value):
        raise NotImplementedError()

    def _triggerModeAuto(self, *args, **kwargs):
        self.sigModeAutoChanged.emit()

    def _triggerCopyActivation(self, *args, **kwargs):
        self.sigCopyActivationChanged.emit()

    def setCopyActive(self, active):
        self._refCopyWidget.setChecked(active)

    def isCopyActive(self):
        return self._refCopyWidget.isChecked()

    def setModeAuto(self, mode_auto):
        old = self.blockSignals(True)
        self._refCopyWidget.setModeAuto(mode_auto)
        self.blockSignals(old)

    def isOnModeAuto(self):
        return self._refCopyWidget.is_on_mode_auto()

    def close(self):
        self.blockSignals(True)
        self._refCopyWidget.close()
        super(DarkRefAndCopyWidget, self).close()

    def _dealWithMissingRef(self, scanID):
        # Security: if lbs is full, skip requesting fir user ref
        if (
            settings.isOnLbsram(scanID)
            and is_low_on_memory(settings.get_lbsram_path()) is True
        ):
            # if computer is running into low memory on lbsram skip it
            mess = (
                "low memory, do not ask user for references (refCopy) "
                "for %s" % scanID
            )
            _logger.processSkipped(mess)
            return
        if (not self._gui.askForFlatUrl()) or (not self._gui.askForDarkUrl()):
            mes = "no reference created for %s, no link registred." % scanID
            _logger.processSkipped(mes)
        else:
            # process ref on this folder if only originals are here
            self.worker.set_process_only_dkRf(True)
            originalFolder = self.worker.directory
            self.worker.directory = self._gui.getCopyFolder()
            self.worker.process()
            self.worker.set_refs_from_scan(self.worker.directory)
            self.worker.directory = originalFolder

            if self.worker.has_flat_or_dark_stored() is True:
                self.worker.set_process_only_copy(True)
                self.worker.run()
            else:
                self._dealWithMissingRef(scanID)


class RefCopyWidget(qt.QGroupBox):
    """
    GUI for the :class:RefCopy
    """

    sigModeAutoChanged = qt.Signal()
    """Signal emitted when the mode auto change"""
    sigCopyActivationChanged = qt.Signal()
    """Signal emitted when the copy is activated or deactivated"""
    sigClearCache = qt.Signal()
    """Signal when the cache needs to be cleared"""

    _DEFAULT_DIRECTORY = "/lbsram/data/visitor"
    """Default directory used when the user need to set path to references"""

    _MSG_NO_REF = "!!! No reference recorded !!!"

    def __init__(self, parent, save_dir: str, refCopy):
        """

        :param parent: Qt parent
        :param save_dir: where the dark and flats 'references' are stored
        """
        qt.QGroupBox.__init__(self, "activate", parent)
        self.__save_dir = save_dir
        self._refCopy = refCopy
        self.setLayout(qt.QVBoxLayout())
        self.setCheckable(True)
        self._infoLabel = qt.QLabel(
            """
            The copy action happen after the reduced (computed) darks and flats. It can be activate or not. \n\n
            Reduced darks and / or flats can be registered and then copy to scan without darks or flat. \n
            You can set manually Darks and flat (button displayed when the mode auto is unchecked.) \n
            Otherwise if the mode 'auto' is activated then each time he meets scan with dark and / or flat it will register those as the one to be copied. \n

            Then each time the copy action meets scan without reduced dark and / or flat it will copy them. \n
            """,
            self,
        )
        self.layout().addWidget(self._infoLabel)
        self.layout().addWidget(self.__createManualGUI())
        self._qcbAutoMode = qt.QCheckBox("auto", parent=self)
        self.layout().addWidget(self._qcbAutoMode)
        spacer = qt.QWidget(self)
        spacer.setSizePolicy(qt.QSizePolicy.Minimum, qt.QSizePolicy.Expanding)
        self.layout().addWidget(spacer)
        self._removeCacheFile = qt.QPushButton("clear cache")
        self._removeCacheFile.setToolTip(
            "Remove the file used for cacheing reduce dark / flat."
        )
        self.layout().addWidget(self._removeCacheFile)

        self.layout().addWidget(self.__createStatusBarGUI())

        self.setModeAuto(True)
        self._qcbAutoMode.toggled.connect(self._updateModeAuto)

        # expose API
        self.sigActivated = self.toggled

        # connect signal / slot
        self.toggled.connect(self._triggerCopyActivated)
        self._qcbAutoMode.toggled.connect(self._triggerModeAutoChanged)
        self._removeCacheFile.released.connect(self.sigClearCache)

    def _triggerCopyActivated(self, *args, **kwargs):
        self.sigCopyActivationChanged.emit()

    def _triggerModeAutoChanged(self, *args, **kwargs):
        self.sigModeAutoChanged.emit()

    def sizeHint(self):
        return qt.QSize(400, 200)

    def __createManualGUI(self):
        self._manualSelectionWidget = qt.QWidget(self)
        self._manualSelectionWidget.setLayout(qt.QFormLayout())

        self._selectDarks = qt.QPushButton("select darks url", parent=self)
        self._manualSelectionWidget.layout().addRow(self._selectDarks)
        self._selectFlats = qt.QPushButton("select flats url", parent=self)
        self._manualSelectionWidget.layout().addRow(self._selectFlats)

        self._cacheFileLabel = qt.QLabel(
            DarkRefsCopy.get_save_file(self.__save_dir), self
        )
        self._manualSelectionWidget.layout().addRow(
            "current cache file is", self._cacheFileLabel
        )
        self._cacheFileLabel.setToolTip(
            "Dark-Flat copy mecanism store darks and flat at a specific location. If you want you can browse the file to see what are the dark and flat you will copy"
        )

        self._selectDarks.released.connect(self.askForDarksUrl)
        self._selectFlats.released.connect(self.askForFlatsUrl)

        return self._manualSelectionWidget

    def __createStatusBarGUI(self):
        self._statusBar = qt.QStatusBar(parent=self)
        self._statusBar.showMessage(self._MSG_NO_REF)
        return self._statusBar

    def _clearRef(self):
        shutil.rmtree(self.__save_dir)

    def askForFlatsUrl(self):
        dialog = DataFileDialog()
        dialog.setDirectory(get_default_directory())
        if dialog.exec():
            url = dialog.selectedUrl()
            try:
                url = DataUrl(path=url)
            except Exception as e:
                _logger.error(f"Fails to define flat url. Error is {e}")
                return False
            else:
                self._statusBar.showMessage(f"darks set from {url.path()}")
                DarkRefsCopy.save_flats_to_be_copied(self.__save_dir, data=url)
                return True
        return False

    def askForDarksUrl(self):
        dialog = DataFileDialog()
        dialog.setDirectory(get_default_directory())
        if dialog.exec():
            url = dialog.selectedUrl()
            try:
                url = DataUrl(path=url)
            except Exception as e:
                _logger.error(f"Fails to define flat url. Error is {e}")
                return False
            else:
                self._statusBar.showMessage(f"darks set from {url.path()}")
                DarkRefsCopy.save_darks_to_be_copied(self.__save_dir, data=url)
                return True
        return False

    def setModeAuto(self, b):
        self._refCopy.set_mode_auto(b)
        self._manualSelectionWidget.setVisible(not b)
        self._qcbAutoMode.setChecked(b)

    def is_on_mode_auto(self):
        return self._qcbAutoMode.isChecked()

    def _updateModeAuto(self):
        """call back of `_qcbGiveManually`"""
        self._refCopy.set_mode_auto(self._qcbAutoMode.isChecked())
        self._manualSelectionWidget.setVisible(not self._qcbAutoMode.isChecked())

    def copyActivated(self):
        """

        :return bool: Return True if the user want to copy reduced darks and flats
        """
        return self.isChecked()

    def save_darks_to_be_copied(self, darks: DataUrl | dict):
        DarkRefsCopy.save_darks_to_be_copied(data=darks, save_dir=self.__save_dir)

    def save_flats_to_be_copied(self, flats: DataUrl | dict):
        DarkRefsCopy.save_flats_to_be_copied(data=flats, save_dir=self.__save_dir)



"""
(Ewoks)Orange Widgets for remote processing through slurm
"""

NAME = "compute cluster"

ID = "orangecontrib.tomwer.widgets.cluster"

DESCRIPTION = "widgets to submit jobs to cluster"

LONG_DESCRIPTION = (
    "widgets to submit slurm job, observe job and convert them back to tomwer object"
)

ICON = "../../widgets/icons/tomwer_cluster.png"

BACKGROUND = "#C0CCFF"

PRIORITY = 5



from __future__ import annotations

from silx.gui import qt
from silx.gui import icons

from silx.gui.widgets.FrameBrowser import FrameBrowser

try:
    from silx.gui.widgets.FrameBrowser import PlayButtonContextMenu
except ImportError:
    from tomwer.third_part.FrameBrowser import PlayButtonContextMenu


class HorizontalSliderWithBrowser(qt.QAbstractSlider):
    """
    Slider widget combining a :class:`QSlider` and a :class:`FrameBrowser`.

    .. image:: img/HorizontalSliderWithBrowser.png

    The data model is an integer within a range.

    The default value is the default :class:`QSlider` value (0),
    and the default range is the default QSlider range (0 -- 99)

    The signal emitted when the value is changed is the usual QAbstractSlider
    signal :attr:`valueChanged`. The signal carries the value (as an integer).

    :param QWidget parent: Optional parent widget
    """

    def __init__(self, parent=None):
        qt.QAbstractSlider.__init__(self, parent)
        self.setOrientation(qt.Qt.Horizontal)

        self.mainLayout = qt.QGridLayout(self)
        self.mainLayout.setContentsMargins(0, 0, 0, 0)
        self.mainLayout.setSpacing(2)

        self._slider = qt.QSlider(self)
        self._slider.setOrientation(qt.Qt.Horizontal)

        self._browser = FrameBrowser(self)

        # avoid focus on the on buttons that can be activated when editing the slice index (from the QLinEdit)
        for button in (
            self._browser.firstButton,
            self._browser.previousButton,
            self._browser.nextButton,
            self._browser.lastButton,
        ):
            button.setFocusPolicy(qt.Qt.FocusPolicy.NoFocus)

        self.mainLayout.addWidget(self._slider, 0, 0, 1, 7)
        self.mainLayout.addWidget(self._browser, 0, 7, 1, 1)

        self._slider.valueChanged[int].connect(self._sliderSlot)
        self._browser.sigIndexChanged.connect(self._browserSlot)

        fontMetric = self.fontMetrics()
        iconSize = qt.QSize(fontMetric.height(), fontMetric.height())

        self.__timer = qt.QTimer(self)
        self.__timer.timeout.connect(self._updateState)

        self._playButton = qt.QToolButton(self)
        self._playButton.setToolTip("Display dataset movie.")
        self._playButton.setIcon(icons.getQIcon("camera"))
        self._playButton.setIconSize(iconSize)
        self._playButton.setCheckable(True)
        self.mainLayout.addWidget(self._playButton, 0, 8, 1, 1)

        self._playButton.toggled.connect(self._playButtonToggled)
        self._menuPlaySlider = PlayButtonContextMenu(self)
        self._menuPlaySlider.sigFrameRateChanged.connect(self._frameRateChanged)
        self._frameRateChanged(self.getFrameRate())
        self._playButton.setMenu(self._menuPlaySlider)
        self._playButton.setPopupMode(qt.QToolButton.MenuButtonPopup)

        # handle slice index buttons
        sliceButtonFont = self.font()
        sliceButtonFont.setPixelSize(8)
        # one-quarters button
        self.mainLayout.setColumnStretch(0, 1)
        self._oneQuarterButton = qt.QPushButton()
        self._oneQuarterButton.setFont(sliceButtonFont)
        self._oneQuarterButton.setFocusPolicy(qt.Qt.FocusPolicy.NoFocus)
        self.mainLayout.addWidget(self._oneQuarterButton, 1, 1, 1, 1)

        # two-quarters button
        self.mainLayout.setColumnStretch(2, 1)
        self._twoQuarterButton = qt.QPushButton()
        self._twoQuarterButton.setFont(sliceButtonFont)
        self._twoQuarterButton.setFocusPolicy(qt.Qt.FocusPolicy.NoFocus)
        self.mainLayout.addWidget(self._twoQuarterButton, 1, 3, 1, 1)

        # three-quarters button
        self.mainLayout.setColumnStretch(4, 1)
        self._threeQuarterButton = qt.QPushButton()
        self._threeQuarterButton.setFont(sliceButtonFont)
        self._threeQuarterButton.setFocusPolicy(qt.Qt.FocusPolicy.NoFocus)
        self.mainLayout.addWidget(self._threeQuarterButton, 1, 5, 1, 1)
        self.mainLayout.setColumnStretch(6, 1)

        self.setSliceIndices(None)
        self.setSliceBrowsingEnabled(False)

        # connect signal / slot
        for button in (
            self._oneQuarterButton,
            self._twoQuarterButton,
            self._threeQuarterButton,
        ):
            button.pressed.connect(self.__quarterButtonPressed)

    def lineEdit(self):
        """Returns the line edit provided by this widget.

        :rtype: qt.QLineEdit
        """
        return self._browser.lineEdit()

    def limitWidget(self):
        """Returns the widget displaying axes limits.

        :rtype: qt.QLabel
        """
        return self._browser.limitWidget()

    def setMinimum(self, value):
        """Set minimum value

        :param int value: Minimum value"""
        self._slider.setMinimum(value)
        maximum = self._slider.maximum()
        self._browser.setRange(value, maximum)

    def setMaximum(self, value):
        """Set maximum value

        :param int value: Maximum value
        """
        self._slider.setMaximum(value)
        minimum = self._slider.minimum()
        self._browser.setRange(minimum, value)

    def setRange(self, first, last):
        """Set minimum/maximum values

        :param int first: Minimum value
        :param int last: Maximum value"""
        self._slider.setRange(first, last)
        self._browser.setRange(first, last)

    def getRange(self) -> tuple[int, int]:
        return self._browser.getRange()

    def _sliderSlot(self, value):
        """Emit selected value when slider is activated"""
        self._browser.setValue(value)
        self.valueChanged.emit(value)

    def _browserSlot(self, ddict):
        """Emit selected value when browser state is changed"""
        self._slider.setValue(ddict["new"])

    def setValue(self, value):
        """Set value

        :param int value: value"""
        self._slider.setValue(value)
        self._browser.setValue(value)

    def value(self):
        """Get selected value"""
        return self._slider.value()

    def setFrameRate(self, value: int):
        """Set the frame rate at which images are displayed"""
        self._menuPlaySlider.setFrameRate(value)

    def getFrameRate(self) -> int:
        """Get the frame rate at which images are displayed"""
        return self._menuPlaySlider.getFrameRate()

    def setPlayImageStep(self, value: int):
        """Set the step between displayed images when playing"""
        self._menuPlaySlider.setStep(value)

    def getPlayImageStep(self) -> int:
        """Returns the step between displayed images"""
        return self._menuPlaySlider.getStep()

    def _frameRateChanged(self, framerate: int):
        """Update the timer interval"""
        self.__timer.setInterval(int(1 / framerate * 1e3))

    def _playButtonToggled(self, checked: bool):
        """Start/Stop the slider sequence."""
        if checked:
            self.__timer.start()
            return
        self.__timer.stop()

    def _updateState(self):
        """Advance an interval number of frames in the browser sequence."""
        currentIndex = self._browser.getValue()
        if currentIndex < self._browser.getRange()[-1]:
            self.setValue(currentIndex + self.getPlayImageStep())
        else:
            self._playButton.setChecked(False)

    def setSliceIndices(self, slices: None | tuple[int]):
        if (slices is not None) and len(slices) != 3:
            raise TypeError("slices is not None or a tuple of three elements.")
        slices_buttons = (
            self._oneQuarterButton,
            self._twoQuarterButton,
            self._threeQuarterButton,
        )
        if slices is None:
            [button.setVisible(False) for button in slices_buttons]
        else:
            slices = sorted(slices)
            for button_index, button in zip(slices, slices_buttons):
                button.setText(str(button_index))
                button.setVisible(True)

    def __quarterButtonPressed(self):
        self._slider.setValue(int(self.sender().text()))

    def setSliceBrowsingEnabled(self, enable):
        """
        Slice browsing is enabled if we agree to let the user browsing the entire stack of slice
        (if the volume is fully load or if on the fast-reading) axis. Else he can only select a pre-defined set of slice.
        """
        self._slider.setEnabled(enable)
        self._playButton.setEnabled(enable)
        self._browser.setEnabled(enable)



# coding: utf-8
from __future__ import annotations

from orangewidget import gui, widget
from orangewidget.widget import Input

from tomwer.gui.debugtools.objectinspector import ObjectInspector


class ObjectInspectorOW(widget.OWBaseWidget, openclass=True):
    """
    A simple widget to browse a TomwerScanBase object
    """

    name = "tomwer object browser"
    id = "orangecontrib.tomwer.widgets.debugtools.tomwerscanbasebrowser"
    description = "create on the fly dataset"
    icon = "icons/inspector.png"
    priority = 255
    keywords = ["tomography", "file", "tomwer", "dataset", "debug"]

    want_main_area = True
    want_control_area = False
    resizing_enabled = True

    class Inputs:
        data = Input(name="object", type=object)

    def __init__(self, parent=None):
        widget.OWBaseWidget.__init__(self, parent=parent)
        self._layout = gui.vBox(self.mainArea, self.name).layout()
        self.inspector = ObjectInspector(parent=self)
        self._layout.addWidget(self.inspector)

    @Inputs.data
    def setObject(self, obj):
        self.inspector.setObject(obj)



# coding: utf-8
from __future__ import annotations

import logging
import os

from orangewidget import gui, widget
from orangewidget.widget import Input

from silx.gui import qt

import tomwer.core.process.visualization.samplemoved
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.gui.samplemoved import SampleMovedWidget

logger = logging.getLogger(__name__)


class SampleMovedOW(widget.OWBaseWidget, openclass=True):
    """
    Simple widget exposing two images side by side to see if a sample moved
    during the acquisition.

    :param parent: the parent widget
    """

    name = "sample moved"
    id = "orange.widgets.tomwer.samplemoved"
    description = (
        "This widget is used to display two scan side by side "
        "to know if a sample moved during the acquisition by "
        "simple observation."
    )
    icon = "icons/sampleMoved.svg"
    priority = 85
    keywords = ["tomography", "sample", "moved", "visualization"]

    ewokstaskclass = (
        tomwer.core.process.visualization.samplemoved._SampleMovedPlaceHolder
    )

    want_main_area = True
    want_control_area = False
    resizing_enabled = True

    class Inputs:
        data = Input(name="data", type=TomwerScanBase)

    def __init__(self, parent=None):
        super().__init__(parent)

        layout = gui.vBox(self.mainArea, self.name).layout()

        self._widgetScanPath = qt.QWidget(parent=self)
        self._widgetScanPath.setLayout(qt.QHBoxLayout())
        self._widgetScanPath.layout().addWidget(
            qt.QLabel("scan: ", parent=self._widgetScanPath)
        )
        self._scanNameQLabel = qt.QLabel("", parent=self._widgetScanPath)
        self._widgetScanPath.layout().addWidget(self._scanNameQLabel)
        spacer = qt.QWidget(parent=self._widgetScanPath)
        spacer.setSizePolicy(qt.QSizePolicy.Expanding, qt.QSizePolicy.Minimum)
        self._widgetScanPath.layout().addWidget(spacer)
        layout.addWidget(self._widgetScanPath)

        self._mainWidget = SampleMovedWidget(parent=self)
        layout.addWidget(self._mainWidget)

    def sizeHint(self):
        return qt.QSize(400, 200)

    @Inputs.data
    def updateScan(self, scan):
        if scan is None:
            return

        assert isinstance(scan, TomwerScanBase)
        if os.path.isdir(scan.path):
            self._scanNameQLabel.setText(os.path.basename(scan.path))
            rawSlices = scan.get_proj_angle_url()

            self._mainWidget.clearOnLoadActions()
            self._mainWidget.setImages(rawSlices)
            self._mainWidget.setScan(scan)
            self._mainWidget.setOnLoadAction(scan.flat_field_correction)

    def clear(self):
        self._scanNameQLabel.setText("")
        self._mainWidget.clear()



import logging

from silx.gui import qt

from tomwer.gui.utils.qt_utils import block_signals
from tomwer.gui.control.actions import TomoObjDisplayModeToolButton
from tomwer.gui.control.tomoobjdisplaymode import DisplayMode

_logger = logging.getLogger(__name__)


class _SelectorWidget(qt.QMainWindow):
    """widget used to select a dataset on a list (a scan or a volume for now)"""

    sigSelectionChanged = qt.Signal(list)
    """Signal emitted when the selection changed. List elements are tomo object identifier as str"""

    sigUpdated = qt.Signal()
    """signal emitted when the scan list change"""

    def __init__(self, parent=None):
        qt.QMainWindow.__init__(self, parent)
        # widget
        self._mainWidget = qt.QWidget(self)
        self._mainWidget.setLayout(qt.QVBoxLayout())
        self.dataList = self._buildDataList()
        self.dataList.setSelectionMode(qt.QAbstractItemView.ExtendedSelection)
        self._mainWidget.layout().addWidget(self.dataList)
        self._mainWidget.layout().addWidget(self._getAddAndRmButtons())
        self._mainWidget.layout().addWidget(self._getSendButtons())
        self._mainWidget.setAcceptDrops(True)
        self.setAcceptDrops(True)
        self.setCentralWidget(self._mainWidget)

        # toolbar
        toolbar = qt.QToolBar()
        toolbar.setMovable(False)
        toolbar.setFloatable(False)
        self.addToolBar(qt.Qt.TopToolBarArea, toolbar)

        tomoObjDisplayAction = TomoObjDisplayModeToolButton(self)
        toolbar.addWidget(tomoObjDisplayAction)

        # set up
        self.setDisplayMode(DisplayMode.SHORT)

        # connect signal / slot
        # for drag / drop we need to have an indirect call to the sigUpdated
        self.dataList.listChanged.connect(self._updatedFromDragDrop)
        tomoObjDisplayAction.sigDisplayModeChanged.connect(self.setDisplayMode)

    def setDisplayMode(self, *args, **kwargs):
        self.dataList.setDisplayMode(*args, **kwargs)

    def _updatedFromDragDrop(self, *args, **kwargs):
        self.sigUpdated.emit()

    def add(self, scan):
        added_scans = self.dataList.add(scan)
        self.sigUpdated.emit()
        return added_scans

    def remove(self, scan):
        self.dataList.remove(scan)
        self.sigUpdated.emit()

    def n_data(self) -> int:
        return len(self.dataList.items)

    def _buildDataList(self):
        raise NotImplementedError("Base class")

    def _getAddAndRmButtons(self):
        lLayout = qt.QHBoxLayout()
        w = qt.QWidget(self)
        w.setLayout(lLayout)
        self._addButton = qt.QPushButton("Add")
        self._addButton.clicked.connect(self._callbackAddData)
        self._rmButton = qt.QPushButton("Remove")
        self._rmButton.clicked.connect(self._callbackRemoveSelectedDatasets)

        spacer = qt.QWidget(self)
        spacer.setSizePolicy(qt.QSizePolicy.Expanding, qt.QSizePolicy.Minimum)
        lLayout.addWidget(spacer)
        lLayout.addWidget(self._addButton)
        lLayout.addWidget(self._rmButton)

        return w

    def _getSendButtons(self):
        layout = qt.QHBoxLayout()
        widget = qt.QWidget(self)
        widget.setLayout(layout)

        # select all button
        self._selectAllButton = qt.QPushButton("Select all")
        self._selectAllButton.setToolTip("Select all scan. (shortcut: Ctrl + A)")
        self._selectAllButton.clicked.connect(self.selectAll)
        layout.addWidget(self._selectAllButton)

        # horizontal spacer
        h_spacer = qt.QWidget(self)
        h_spacer.setSizePolicy(qt.QSizePolicy.Expanding, qt.QSizePolicy.Minimum)
        layout.addWidget(h_spacer)

        # send selected button
        self._sendButton = qt.QPushButton("Send selected")
        self._sendButton.clicked.connect(self._sendSelected)
        layout.addWidget(self._sendButton)

        return widget

    def selectAll(self):
        self.dataList.selectAll()

    def setMySelection(self, selection: tuple):
        self.dataList.setMySelection(selection)

    def _callbackAddData(self):
        raise NotImplementedError("Base class")

    def _sendSelected(self):
        sItem = self.dataList.selectedItems()
        if sItem and len(sItem) >= 1:
            selection = [
                _item.data(qt.Qt.UserRole).get_identifier().to_str() for _item in sItem
            ]
            self.sigSelectionChanged.emit(list(selection))
        else:
            _logger.warning("No active scan detected")

    def _sendAll(self):
        allItems = list(self.dataList._myitems.keys())
        if len(allItems) > 0:
            self.sigSelectionChanged.emit(allItems)

    def _callbackRemoveSelectedDatasets(self):
        """ """
        selected_items = self.dataList.selectedItems()
        tomwer_objs_to_remove = [
            item.data(qt.Qt.UserRole).get_identifier().to_str()
            for item in selected_items
        ]
        for tomwer_obj in tomwer_objs_to_remove:
            self.remove(tomwer_obj)
        self.sigUpdated.emit()

    def setActiveData(self, data):
        """
        set the given scan as the active one

        :param scan: the scan to set active
        """
        data_id = data
        self.dataList.setCurrentItem(self.dataList.items[data_id])

    def removeSelectedDatasets(self):
        sItem = self.dataList.selectedItems()
        selection = [
            _item.data(qt.Qt.UserRole).get_identifier().to_str() for _item in sItem
        ]

        with block_signals(self):
            # make sure sigUpdated is called only once.
            for identifier in selection:
                self.remove(identifier)
        self.sigUpdated.emit()



class BaseResourceObserver:
    """base class of the resource observer"""

    def release_resource(self, resource):
        raise NotImplementedError("Base class")


class _DummyResourceManager:
    """
    Simple resource manager. It holds a set of observer tha must release a resource when requested
    Observers must implement the 'BaseResourceObserver'
    """

    def __init__(self):
        self._observers = set()

    def register(self, observer):
        """Register a process to be notified when the resource should be released."""
        if observer not in self._observers:
            self._observers.add(observer)

    def unregister(self, observer):
        """Unregister a process so it won't be notified anymore."""
        if observer in self._observers:
            self._observers.remove(observer)

    def release_resource(self, resource: str):
        """Force all observer to release the given resource"""
        for observer in self._observers:
            observer.release_resource(resource)


HDF5VolumeManager = _DummyResourceManager()



# coding: utf-8
"""
data watcher classes used to define the status of an acquisition for EDF
acquisitions
"""
from __future__ import annotations

import logging

import h5py

from .datawatcherprocess import _DataWatcherProcess

_logger = logging.getLogger(__name__)


class _DataWatcherProcessHDF5(_DataWatcherProcess):
    """
    look for hdf5 information
    """

    def __init__(self, dataDir, srcPattern, destPattern):
        super(_DataWatcherProcessHDF5, self).__init__(dataDir, srcPattern, destPattern)
        if h5py.is_hdf5(dataDir):
            self._nxtomo_file = dataDir

    def _removeAcquisition(self, scanID, reason):
        _logger.warning(
            "removing acquisition is not done for hdf5 data watcher " "process"
        )

    def is_data_complete(self):
        # For now data complete and is abort_is not handled for hdf5 and Nxtomo
        return True

    def is_abort(self):
        # For now data complete and is abort_is not handled for hdf5 and Nxtomo
        return False


class _BlissScanWatcherProcess(_DataWatcherProcess):
    def __init__(self, dataDir, srcPattern=None, destPattern=None):
        super().__init__(dataDir, srcPattern, destPattern)
        if h5py.is_hdf5(dataDir):
            self._blissScanFile = dataDir

    def _removeAcquisition(self, scanID, reason):
        _logger.warning(
            "remoing acquisition is not done for hdf5 data watcher " "process"
        )

    def is_data_complete(self):
        # For now data complete and is abort_is not handled for hdf5 and Nxtomo
        return True

    def is_abort(self):
        # For now data complete and is abort_is not handled for hdf5 and Nxtomo
        return False



from ewokscore.task import Task as EwoksTask
from tomwer.core.utils.scanutils import data_identifier_to_scan


class _DataDiscoveryPlaceHolder(EwoksTask, input_names=["data"], output_names=["data"]):
    """
    Task to recursivly search under a root folder for any scan / data.
    """

    def run(self):
        self.outputs.data = data_identifier_to_scan(self.inputs.data)



# coding: utf-8
from __future__ import annotations


import functools
import logging

from processview.core.manager import DatasetState, ProcessManager
from processview.core.superviseprocess import SuperviseProcess
from silx.gui import qt

from tomwer.core.process.reconstruction.normalization import SinoNormalizationTask
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.settings import get_lbsram_path, isOnLbsram
from tomwer.core.utils.lbsram import is_low_on_memory

from ..processingstack import FIFO, ProcessingThread

_logger = logging.getLogger(__name__)


class INormalizationProcessStack(FIFO, qt.QObject):
    """Implementation of the `.AxisTask` but having a stack for treating
    scans and making computation in threads"""

    def __init__(self, process_id=None):
        qt.QObject.__init__(self)
        FIFO.__init__(self, process_id=process_id)

    def _process(self, data, configuration, callback=None):
        ProcessManager().notify_dataset_state(
            dataset=data,
            process=self,
            state=DatasetState.ON_GOING,
        )
        _logger.processStarted(f"start intensity normalization {data}")
        assert isinstance(data, TomwerScanBase)
        if isOnLbsram(data) and is_low_on_memory(get_lbsram_path()) is True:
            # if computer is running into low memory on lbsram skip it
            mess = "low memory, skip intensity normalization", data.path
            ProcessManager().notify_dataset_state(
                dataset=data, process=self._process_id, state=DatasetState.SKIPPED
            )
            _logger.processSkipped(mess)
            if callback is not None:
                callback()
            self.scan_ready(scan=data)
        else:
            self._data_currently_computed = data
            self._computationThread.init(data=data, i_norm_params=configuration)
            # need to manage connect before starting it because
            fct_callback = functools.partial(self._end_threaded_computation, callback)
            self._computationThread.finished.connect(fct_callback)
            self._computationThread.start()

    def _end_threaded_computation(self, callback=None):
        assert self._data_currently_computed is not None
        self._computationThread.finished.disconnect()
        if callback:
            callback()
        FIFO._end_threaded_computation(self)

    def _create_processing_thread(self, process_id=None) -> qt.QThread:
        return _ProcessingThread(process_id=process_id)


class _ProcessingThread(ProcessingThread, SuperviseProcess):
    """
    Thread use to execute the processing of intensity normalization
    (mostly load io and compute the mean / median on it)
    """

    def __init__(self, process_id=None):
        SuperviseProcess.__init__(self, process_id=process_id)
        ProcessingThread.__init__(self, process_id=process_id)
        self.center_of_rotation = None
        self._dry_run = False
        self._scan = None
        self._i_norm_params = None

    def init(self, data, i_norm_params):
        self._scan = data
        self._i_norm_params = i_norm_params

    def run(self):
        self.sigComputationStarted.emit()
        norm_process = SinoNormalizationTask(
            process_id=self.process_id,
            varinfo=None,
            inputs={
                "data": self._scan,
                "configuration": self._i_norm_params,
                "serialize_output_data": False,
            },
        )
        try:
            norm_process.run()
        except Exception as e:
            _logger.error(str(e))
            mess = f"Intensity normalization computation for {self._scan} failed."
            state = DatasetState.FAILED
        else:
            mess = f"Intensity normalization computation for {self._scan} succeed."
            state = DatasetState.WAIT_USER_VALIDATION

        ProcessManager().notify_dataset_state(
            dataset=self._scan,
            process=self,
            state=state,
            details=mess,
        )



from __future__ import annotations

import os
import numpy
from .dracbase import DracDatasetBase
from tomoscan.esrf.volume.singleframebase import VolumeSingleFrameBase
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.volume.volumebase import TomwerVolumeBase
from tomwer.core.process.reconstruction.nabu.plane import NabuPlane
from tomwer.core.process.reconstruction.output import (
    PROCESS_FOLDER_RECONSTRUCTED_VOLUMES,
)
from tomwer.core.process.reconstruction.output import (
    PROCESS_FOLDER_CAST_VOLUME,
)


__all__ = ["DracReconstructedVolumeDataset"]


class DracReconstructedVolumeDataset(DracDatasetBase):
    """
    Class to associate reconstructed volume(s) to an drac (processed) dataset
    """

    def __init__(self, tomo_obj: TomwerVolumeBase, source_scan: TomwerScanBase) -> None:
        if not isinstance(tomo_obj, TomwerVolumeBase):
            raise TypeError(
                f"tomo_obj should be an instance of {TomwerVolumeBase}. Got {type(tomo_obj)}"
            )

        super().__init__(tomo_obj=tomo_obj, data_dir=tomo_obj.icat_data_dir)
        self._n_slices_per_axis = 3
        self.__source_scan = source_scan
        self.__bliss_raw_dataset = self.from_scan_to_raws(scan=self.__source_scan)

    @staticmethod
    def make_serializable(obj):
        """
        Recursively convert numpy arrays to lists and replace string "None" with None in a given object.
        """
        if isinstance(obj, dict):
            return {
                key: DracReconstructedVolumeDataset.make_serializable(value)
                for key, value in obj.items()
            }
        elif isinstance(obj, (list, tuple)):
            return [
                DracReconstructedVolumeDataset.make_serializable(item) for item in obj
            ]
        elif isinstance(obj, numpy.ndarray):
            return obj.tolist()
        else:
            return obj

    @property
    def metadata(self) -> dict:
        assert isinstance(self.tomo_obj, TomwerVolumeBase)
        metadata = self.tomo_obj.build_drac_metadata()
        metadata["Sample_name"] = self.source_scan.sample_name
        return metadata

    @property
    def dataset_name(self) -> str:
        """name to give to the drac (processed) dataset."""
        return PROCESS_FOLDER_RECONSTRUCTED_VOLUMES

    @property
    def source_scan(self) -> TomwerScanBase:
        return self.__source_scan

    @property
    def bliss_raw_datasets(self) -> tuple[str]:
        return self.__bliss_raw_dataset

    @property
    def n_slices_per_axis(self) -> int:
        "number of slices to sample per axis"
        return self._n_slices_per_axis

    @n_slices_per_axis.setter
    def n_slices_per_axis(self, n: int) -> None:
        self._n_slices_per_axis = n

    def get_slices_to_extract(self) -> tuple[tuple[int, tuple[int]]]:
        """
        Compute the slices to be retrieve along each dimension according to 'n_slices_per_axis'

        return tuple (A) is a two elements tuple. First element if the axis (B).
        Second is the tuple of indices to extract along the axis (B)
        indices are equally spaced in each dimensions
        """
        result: list[tuple[int, tuple[int]]] = []
        volume = self.tomo_obj
        if not isinstance(volume, TomwerVolumeBase):
            raise TypeError(
                f"Volume is expected to be an instance of {TomwerVolumeBase}. Got {type(volume)}"
            )
        volume_shape = volume.get_volume_shape()
        for axis, axis_len in enumerate(volume_shape):
            for slice_index in numpy.linspace(
                0, axis_len, endpoint=False, num=self.n_slices_per_axis + 1
            )[1:]:
                result.append((axis, numpy.round(slice_index).astype(numpy.uint16)))
        return tuple(result)

    def build_gallery(self):
        gallery_dir = self.get_gallery_dir()
        volume = self.tomo_obj
        slices = volume.get_slices(slices=self.get_slices_to_extract())

        # Stack all slices into a single array to compute the global percentiles
        all_slices = numpy.concatenate([slice_ for slice_ in slices.values()])
        lower_bound = numpy.percentile(all_slices, 0)
        upper_bound = numpy.percentile(all_slices, 100)

        for (axis, slice_index), slice_ in slices.items():
            self.save_to_gallery(
                output_file_name=self.get_output_file_name(
                    output_dir=gallery_dir,
                    axis=axis,
                    slice_index=slice_index,
                    volume=volume,
                ),
                image=slice_,
                lower_bound=lower_bound,
                upper_bound=upper_bound,
            )

    @staticmethod
    def get_output_file_name(
        output_dir: str, axis: int, slice_index: int, volume: TomwerVolumeBase
    ) -> str:
        axis = NabuPlane.from_value(axis)
        if isinstance(volume, VolumeSingleFrameBase):
            basename = volume.get_volume_basename()
        else:
            basename = os.path.splitext(os.path.basename(volume.data_url.file_path()))[
                0
            ]
        axis_folder = os.path.join(
            os.path.abspath(output_dir),
            str(axis.value),
        )
        os.makedirs(axis_folder, exist_ok=True)
        # warning: for ordering drac needs to have the axis in a dedicated folder
        return os.path.join(
            axis_folder,
            f"{basename}_capture_{axis.value}_{str(slice_index).zfill(6)}",
        )


class DracCastReconstructedVolume(DracReconstructedVolumeDataset):
    """
    Class to associate casted - reconstructed volume(s) to an drac (processed) dataset
    """

    @property
    def dataset_name(self) -> str:
        """name to give to the drac (processed) dataset."""
        return PROCESS_FOLDER_CAST_VOLUME



import numpy
from silx.gui.utils.testutils import SignalListener
from tomwer.tests.conftest import qtapp  # noqa F401
from tomwer.gui.control.reducedarkflatselector import (
    filter_selected_reduced_frames,
    filter_unselected_reduced_frames,
    ReduceDarkFlatSelectorWidget,
    ReduceDarkFlatSelectorDialog,
)


def test_reduce_darkflat_selector(
    qtapp,  # noqa F811
):
    widget = ReduceDarkFlatSelectorWidget()
    widget.show()

    widget.addReduceFrames(
        {
            0: numpy.random.random(100 * 100).reshape(100, 100),
            1: numpy.random.random(100 * 100).reshape(100, 100),
        }
    )

    widget.addReduceFrames(
        {
            "reduce_frames_name": "my_dict",
            -1: numpy.random.random(100 * 100).reshape(100, 100),
            2: numpy.random.random(100 * 100).reshape(100, 100),
        }
    )

    config = widget.getConfiguration()
    assert len(config) == 2
    assert config[0]["reduce_frames_name"] == "reduced frames #0"
    assert config[1]["reduce_frames_name"] == "my_dict"

    widget.setConfiguration(
        (
            {
                "reduce_frames_name": "toto frames",
                "reduce_frames": (
                    {
                        "index": -1,  # -1 will be interpreated as at the end of the acquisition
                        "data": numpy.random.random(20 * 20).reshape(20, 20),
                        "selected": True,
                    },
                    {
                        "index": 0,
                        "data": numpy.random.random(20 * 20).reshape(20, 20),
                        "selected": True,
                    },
                ),
            },
            {
                "reduce_frames_name": "tata frames",
                "reduce_frames": (
                    {
                        "index": "0.25r",
                        "data": numpy.random.random(40 * 20).reshape(40, 20),
                        "selected": False,
                    },
                ),
            },
        )
    )

    config = widget.getConfiguration()
    assert len(config) == 2
    # check selection has been well take into account
    assert config[0]["reduce_frames"][0]["selected"]
    assert config[0]["reduce_frames"][1]["selected"]
    assert not config[1]["reduce_frames"][0]["selected"]

    selected_frames = widget.getSelectedReduceFrames()
    assert -1 in selected_frames
    assert 0 in selected_frames
    assert 0.25 not in selected_frames

    # test receive twice the same 'reduce_frames_name'
    widget.clear()
    assert len(widget.getConfiguration()) == 0
    for i in range(2):
        widget.addReduceFrames(
            {
                "reduce_frames_name": "my_dict",
                -1: numpy.random.random(100 * 100).reshape(100, 100),
                2: numpy.random.random(100 * 100).reshape(100, 100),
            }
        )
    assert len(widget.getConfiguration()) == 1


def test_reduce_darkflat_selector_dialog(
    qtapp,  # noqa F811
):
    dialog = ReduceDarkFlatSelectorDialog()

    flat_sel_signal_listener = SignalListener()
    dialog.sigSelectActiveAsFlats.connect(flat_sel_signal_listener)

    dark_sel_signal_listener = SignalListener()
    dialog.sigSelectActiveAsDarks.connect(dark_sel_signal_listener)

    dialog.show()

    dialog.addReduceFrames(
        {
            0: numpy.random.random(100 * 100).reshape(100, 100),
            1: numpy.random.random(100 * 100).reshape(100, 100),
        },
        selected=(1,),
    )

    dialog.addReduceFrames(
        {
            "reduce_frames_name": "my_dict",
            -1: numpy.random.random(100 * 100).reshape(100, 100),
            2: numpy.random.random(100 * 100).reshape(100, 100),
        },
        selected=(-1,),
    )

    # check dark selection
    assert dark_sel_signal_listener.callCount() == 0
    dialog._darkSelected()
    assert dark_sel_signal_listener.callCount() == 1
    assert -1 in dark_sel_signal_listener.arguments()[0][0]

    # check flat selection
    assert flat_sel_signal_listener.callCount() == 0
    dialog._flatSelected()
    assert flat_sel_signal_listener.callCount() == 1
    assert -1 in flat_sel_signal_listener.arguments()[0][0]

    # check remove selected
    dialog._removeSelected()
    assert dialog._widget.getSelectedReduceFrames() == {}

    # check clear selection
    dialog.addReduceFrames(
        {
            "reduce_frames_name": "new dict",
            -1: numpy.random.random(100 * 100).reshape(100, 100),
            2: numpy.random.random(100 * 100).reshape(100, 100),
        },
        selected=(-1, 2),
    )

    assert len(dialog._widget.getSelectedReduceFrames()) == 2
    dialog._clearSelection()
    assert len(dialog._widget.getSelectedReduceFrames()) == 0


def test_reduced_frames_filtering():
    """
    dummy test of the filter_selected_reduced_frames filter_unselected_reduced_frames function
    """
    check_reduce_frames_configuration(
        filter_selected_reduced_frames(
            (
                {
                    "reduce_frames_name": "toto frames",
                    "reduce_frames": (
                        {
                            "index": -1,  # -1 will be interpreated as at the end of the acquisition
                            "data": numpy.random.random(20 * 20).reshape(20, 20),
                            "selected": True,
                        },
                        {
                            "index": 0,
                            "data": numpy.random.random(20 * 20).reshape(20, 20),
                            "selected": True,
                        },
                    ),
                },
                {
                    "reduce_frames_name": "tata frames",
                    "reduce_frames": (
                        {
                            "index": "0.25r",
                            "data": numpy.random.random(40 * 20).reshape(40, 20),
                            "selected": False,
                        },
                    ),
                },
            )
        ),
        (
            {
                "reduce_frames_name": "toto frames",
                "reduce_frames": (
                    {
                        "index": -1,  # -1 will be interpreated as at the end of the acquisition
                        "data": numpy.random.random(20 * 20).reshape(20, 20),
                        "selected": True,
                    },
                    {
                        "index": 0,
                        "data": numpy.random.random(20 * 20).reshape(20, 20),
                        "selected": True,
                    },
                ),
            },
        ),
    )

    check_reduce_frames_configuration(
        filter_unselected_reduced_frames(
            (
                {
                    "reduce_frames_name": "toto frames",
                    "reduce_frames": (
                        {
                            "index": -1,  # -1 will be interpreated as at the end of the acquisition
                            "data": numpy.random.random(20 * 20).reshape(20, 20),
                            "selected": True,
                        },
                        {
                            "index": 0,
                            "data": numpy.random.random(20 * 20).reshape(20, 20),
                            "selected": False,
                        },
                    ),
                },
                {
                    "reduce_frames_name": "tata frames",
                    "reduce_frames": (
                        {
                            "index": "0.25r",
                            "data": numpy.random.random(40 * 20).reshape(40, 20),
                            "selected": False,
                        },
                    ),
                },
            )
        ),
        (
            {
                "reduce_frames_name": "toto frames",
                "reduce_frames": (
                    {
                        "index": 0,
                        "data": numpy.random.random(20 * 20).reshape(20, 20),
                        "selected": False,
                    },
                ),
            },
            {
                "reduce_frames_name": "tata frames",
                "reduce_frames": (
                    {
                        "index": "0.25r",
                        "data": numpy.random.random(40 * 20).reshape(40, 20),
                        "selected": False,
                    },
                ),
            },
        ),
    )

    check_reduce_frames_configuration(filter_unselected_reduced_frames(()), ())


def check_reduce_frames_configuration(config_1, config_2):
    """
    simple comparaison of two configurations
    """
    assert len(config_1) == len(config_2)
    for reduce_frame_group_1, reduce_frame_group_2 in zip(config_1, config_2):
        assert reduce_frame_group_1.get(
            "reduce_frames_name", None
        ) == reduce_frame_group_2.get("reduce_frames_name", None)

        for reduce_frame_group_1, reduce_frame_group_2 in zip(
            reduce_frame_group_1.get("reduce_frames", tuple()),
            reduce_frame_group_2.get("reduce_frames", tuple()),
        ):
            assert reduce_frame_group_1["index"] == reduce_frame_group_2["index"]
            assert reduce_frame_group_1["selected"] == reduce_frame_group_2["selected"]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Application to display findable reconstructed slice of a scan (as a stack).
reconstructed slices can be added by hand or as input(s) of the command line.
"""

import argparse
import logging
import signal
import sys
import os

import silx
from silx.gui import qt

from tomwer.core.utils.resource import increase_max_number_file
from tomwer.gui import icons
from tomwer.gui.stacks import SliceStack
from tomwer.gui.utils.splashscreen import getMainSplashScreen

logging.basicConfig(level=logging.WARNING)
_logger = logging.getLogger(__name__)


def getinputinfo():
    return "tomwer slice-stack [scanDir]"


def sigintHandler(*args):
    """Handler for the SIGINT signal."""
    qt.QApplication.quit()


def main(argv):
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "volume_paths",
        help="Volumes to be added to the slice stack visualization. Warning: volumes should be single slice volumes.",
        nargs="*",
    )
    parser.add_argument(
        "--use-opengl-plot",
        "--opengl-backend",
        help="Use OpenGL for plots (instead of matplotlib)",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--debug",
        dest="debug",
        action="store_true",
        default=False,
        help="Set logging system in debug mode",
    )

    options = parser.parse_args(argv[1:])

    if options.use_opengl_plot:
        silx.config.DEFAULT_PLOT_BACKEND = "gl"
    else:
        silx.config.DEFAULT_PLOT_BACKEND = "matplotlib"

    increase_max_number_file()

    global app  # QApplication must be global to avoid seg fault on quit
    app = qt.QApplication.instance() or qt.QApplication(["tomwer"])

    qt.QLocale.setDefault(qt.QLocale(qt.QLocale.English))
    qt.QLocale.setDefault(qt.QLocale.c())
    signal.signal(signal.SIGINT, sigintHandler)
    sys.excepthook = qt.exceptionHandler
    timer = qt.QTimer()
    timer.start(500)
    # Application have to wake up Python interpreter, else SIGINT is not
    # catched
    timer.timeout.connect(lambda: None)

    splash = getMainSplashScreen()
    widget = SliceStack()
    widget.setWindowIcon(icons.getQIcon("tomwer"))
    widget._addTomoObjectsFromStrList(
        [os.path.abspath(volume_path) for volume_path in options.volume_paths]
    )

    splash.finish(widget)
    widget.show()
    app.exec()


if __name__ == "__main__":
    main(sys.argv)



# coding: utf-8
from __future__ import annotations


import logging
import time
import weakref
from packaging.version import Version
from contextlib import AbstractContextManager

from silx.gui import qt

from tomwer.core.process.control.scanvalidator import ScanValidator as BaseDataValidator
from tomwer.core.scan.scanbase import TomwerScanBase, _TomwerBaseDock
from tomwer.core.scan.scanfactory import ScanFactory
from tomwer.gui.dialog.QDataDialog import QDataDialog
from tomwer.gui.utils.waiterthread import QWaiterThread
from tomwer.gui.visualization.dataviewer import DataViewer

_logger = logging.getLogger(__file__)


WAIT_TIME_MEM_REL = 20
"""Time (in sec) to wait to check if the scan validator has to release his
stack of scans. This is a security to make sure we are not keeping
unnecessary data during an acquisition in lbsram"""


class DataValidator(qt.QMainWindow, BaseDataValidator):
    """
    Widget used to validate an acquisition or recall some processing
    """

    sigScanReady = qt.Signal(TomwerScanBase)
    """signal emitted when a scan is validated"""

    sigChangeReconsParams = qt.Signal(_TomwerBaseDock)
    """signal emitted when the user request modifications on reconstruction
    parameters"""

    class SliderCM(AbstractContextManager):
        """Simple context manager to hida / show button dialogs"""

        def __init__(self, slider):
            self._slider = slider
            self._old = None

        def __enter__(self):
            self._old = self._slider.blockSignals(True)

        def __exit__(self, exc_type, exc_val, exc_tb):
            self._slider.blockSignals(self._old)

    def __init__(self, parent, timeout_loop_memory=WAIT_TIME_MEM_REL):
        qt.QMainWindow.__init__(
            self, parent, memoryReleaser=QWaiterThread(timeout_loop_memory)
        )
        # needed on lbs191... otherwise crash with PyQt5.11. with a QThread destroyed...
        time.sleep(0.01)
        BaseDataValidator.__init__(
            self, memoryReleaser=QWaiterThread(timeout_loop_memory)
        )
        self._centralWidget = DataViewer(parent=parent)
        self.setCentralWidget(self._centralWidget)

        # scan slider and add scan
        self._lateralWidget = qt.QWidget(self)
        self._lateralWidget.setLayout(qt.QVBoxLayout())
        self._addScanButton = qt.QPushButton(self)
        style = qt.QApplication.style()
        icon = style.standardIcon(qt.QStyle.SP_DirIcon)
        self._addScanButton.setIcon(icon)
        self._lateralWidget.layout().addWidget(self._addScanButton)
        self._scanSelectorSlider = _ScanSelectorWidget(parent=self)
        self._lateralWidget.layout().addWidget(self._scanSelectorSlider)
        self._lateralWidget.layout().setContentsMargins(0, 0, 0, 0)

        self._scanSelectorSliderDocker = qt.QDockWidget(self)
        self._scanSelectorSliderDocker.setWidget(self._lateralWidget)
        self._scanSelectorSliderDocker.setFeatures(qt.QDockWidget.DockWidgetMovable)
        self.addDockWidget(qt.Qt.RightDockWidgetArea, self._scanSelectorSliderDocker)

        # control buttons
        self._controlButtons = _DataControlWidget(parent=self)
        self._controlButtonsDocker = qt.QDockWidget(self)
        self._controlButtonsDocker.setWidget(self._controlButtons)
        self._controlButtonsDocker.setFeatures(qt.QDockWidget.DockWidgetMovable)
        self.addDockWidget(qt.Qt.BottomDockWidgetArea, self._controlButtonsDocker)

        self._scanSelectorSlider.setMinimum(0)
        self._scanSelectorSlider.setMaximum(0)

        # connect signal / slot
        self._scanSelectorSlider.sigScanChanged.connect(self._updateCurrentScan)
        _controlButtons = self._controlButtons
        _controlButtons.sigValidateScan.connect(self.validateCurrentScan)
        _controlButtons.sigChangeReconstructionParametersScan.connect(
            self.changeReconsParamCurrentScan
        )
        _controlButtons.toggled.connect(self.setManualValidation)
        self._addScanButton.pressed.connect(self._addNewScanCallback)

    def close(self) -> bool:
        self._centralWidget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self._centralWidget.close()
        self._centralWidget = None
        qt.QMainWindow.close(self)

    def addScan(self, scan):
        """

        :param scan:
        :return:
        """
        assert isinstance(scan, TomwerScanBase)
        BaseDataValidator.addScan(self, scan)

        with self.SliderCM(self._scanSelectorSlider):
            current_index = self._scanSelectorSlider.value()
            self._scanSelectorSlider.setRange(0, len(self._scans) - 1)
            self._scanSelectorSlider.setValue(current_index)

        if str(scan) in self._scans:
            self.setActiveScan(scan)

    def setActiveScan(self, scan: TomwerScanBase):
        """

        :param scan:
        :return:
        """
        if scan is None:
            self.clear()
        elif str(scan) not in self._scans:
            _logger.warning("scan has not been registered")
        else:
            with self.SliderCM(self._scanSelectorSlider):
                self._centralWidget.setScan(scan)
                self._controlButtons.setScan(scan)
                index = list(self._scans.keys()).index(str(scan))
                self._scanSelectorSlider.setValue(index)

    def _updateCurrentScan(self):
        with self.SliderCM(self._scanSelectorSlider):
            self.setActiveScan(self.getCurrentScan())

    def getCurrentScan(self) -> TomwerScanBase:
        """

        :return:
        """
        current_scan_index = self._scanSelectorSlider.value()
        if current_scan_index >= len(self._scans):
            return None
        else:
            scan_id = list(self._scans.keys())[current_scan_index]
            scan = self._scans[scan_id]
            assert isinstance(scan, TomwerScanBase)
            return scan

    def validateCurrentScan(self):
        scan = self.getCurrentScan()
        if scan is None:
            return
        else:
            BaseDataValidator._validateScan(self, scan)
            with self.SliderCM(self._scanSelectorSlider):
                current_index = self._scanSelectorSlider.value()
                self._scanSelectorSlider.setRange(0, len(self._scans) - 1)
                # note: slider force the value to be in [minimum, maximum]
                self._scanSelectorSlider.setValue(current_index)
                self._updateCurrentScan()

    def changeReconsParamCurrentScan(self):
        current_scan = self.getCurrentScan()
        if current_scan:
            assert isinstance(current_scan, TomwerScanBase)
            BaseDataValidator._changeReconsParam(self, current_scan)

    def _sendScanReady(self, scan):
        if scan is not None:
            assert isinstance(scan, TomwerScanBase)
            self.sigScanReady.emit(scan)

    def _sendUpdateReconsParam(self, scan):
        if scan is not None:
            assert isinstance(scan, _TomwerBaseDock)
            self.sigChangeReconsParams.emit(scan)

    def clear(self):
        try:
            self._centralWidget.clear()
        except AttributeError:
            # if has been removed before calling clear
            pass

    def _validateStack(self, filter_=None):
        BaseDataValidator._validateStack(self, filter_=filter_)
        self._updateCurrentScan()

    def setAutomaticValidation(self, auto):
        self._controlButtons.setChecked(not auto)

    def _addNewScanCallback(self):
        """Callback when we want to add a new scan to the scan validator"""
        dialog = QDataDialog(self, multiSelection=True)

        if not dialog.exec():
            dialog.close()
            return

        foldersSelected = dialog.files_selected()
        new_active_scan = None
        for folder in foldersSelected:
            try:
                scans = ScanFactory.create_scan_objects(scan_path=folder)
            except Exception as e:
                _logger.error(
                    f"cannot create scan instances from {folder}. Error is {e}"
                )
            else:
                for scan in scans:
                    self.addScan(scan=scan)
                    new_active_scan = scan

        if new_active_scan is not None:
            self.setActiveScan(new_active_scan)


class _ScanSelectorWidget(qt.QWidget):
    """Slider to select the scan we want to 'check'"""

    sigScanChanged = qt.Signal(int)
    """signal emitted when the scan change. Value is the scan index"""

    def __init__(self, parent):
        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QHBoxLayout())

        self._qslider = qt.QSlider(qt.Qt.Vertical, self)
        self.layout().addWidget(self._qslider)

        self._verticalLabel = _VerticalLabel(
            "stack of received scan", parent=self, revert=True
        )
        self.layout().addWidget(self._verticalLabel)

        # connect signal / slot
        self._qslider.valueChanged.connect(self._valueHasChanged)

    def _valueHasChanged(self, *args, **kwargs):
        self.sigScanChanged.emit(self._qslider.value())

    def value(self):
        return self._qslider.value()

    def setValue(self, value):
        self._qslider.setValue(value)

    def setMinimum(self, p_int):
        self._qslider.setMinimum(p_int)

    def setMaximum(self, p_int):
        self._qslider.setMaximum(p_int)

    def setRange(self, p_int, p_int_1):
        self._qslider.setRange(p_int, p_int_1)


class _DataControlWidget(qt.QGroupBox):
    """Set of buttons to control the current scan"""

    sigValidateScan = qt.Signal(str)
    sigCancelScan = qt.Signal(str)
    sigRedoAcquisitionScan = qt.Signal(TomwerScanBase)
    sigChangeReconstructionParametersScan = qt.Signal(TomwerScanBase)

    def __init__(self, parent=None):
        qt.QGroupBox.__init__(self, title="Validate manually", parent=parent)
        self._scan = None
        self.setCheckable(True)
        layout = qt.QGridLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)

        #  validate button
        self.validateButton = qt.QPushButton("Validate")
        style = qt.QApplication.style()
        self.validateButton.setIcon(style.standardIcon(qt.QStyle.SP_DialogApplyButton))
        self.validateButton.pressed.connect(self.__validated)
        layout.addWidget(self.validateButton, 0, 2)

        spacer = qt.QWidget(self)
        spacer.setSizePolicy(qt.QSizePolicy.Expanding, qt.QSizePolicy.Minimum)
        layout.addWidget(spacer, 0, 1)

        #  change reconstruction parameters button
        self.changeReconsParamButton = qt.QPushButton(
            "Change reconstruction parameters"
        )
        self.changeReconsParamButton.setIcon(
            style.standardIcon(qt.QStyle.SP_FileDialogContentsView)
        )
        self.changeReconsParamButton.pressed.connect(
            self.__updateReconstructionParameters
        )
        layout.addWidget(self.changeReconsParamButton, 2, 0)

    def setScan(self, scan):
        self._scan = weakref.ref(scan)

    def setEnabled(self, b):
        self.validateButton.setEnabled(b)
        self.changeReconsParamButton.setEnabled(b)

    def __validated(self):
        """Callback when the validate button is pushed"""
        self.sigValidateScan.emit("")

    def __redoacquisition(self):
        """Callback when the redo acquisition button is pushed"""
        if self._scan is not None and self._scan() is not None:
            self.sigRedoAcquisitionScan.emit(self._scan())

    def __updateReconstructionParameters(self):
        """Callback when the change reconstruction button is pushed"""
        if self._scan is not None and self._scan() is not None:
            self.sigChangeReconstructionParametersScan.emit(self._scan())


class _VerticalLabel(qt.QLabel):
    """Display vertically the given text"""

    def __init__(self, text, parent=None, revert=False):
        """

        :param text: the legend
        :param parent: the Qt parent if any
        """
        qt.QLabel.__init__(self, text, parent)
        self.revert = revert
        self.setLayout(qt.QVBoxLayout())
        self.layout().setContentsMargins(0, 0, 0, 0)

    def paintEvent(self, event):
        painter = qt.QPainter(self)
        painter.setFont(self.font())

        painter.translate(0, self.rect().height())
        painter.rotate(90)
        if self.revert:
            newRect = qt.QRect(
                -self.rect().height(),
                -self.rect().width(),
                self.rect().height(),
                self.rect().width(),
            )
        else:
            newRect = self.rect()

        painter.drawText(newRect, qt.Qt.AlignHCenter, self.text())

        fm = qt.QFontMetrics(self.font())
        if Version(qt.QT_VERSION_STR) >= Version("5.11"):
            preferedHeight = fm.horizontalAdvance(self.text())
        else:
            preferedHeight = fm.width(self.text())
        preferedWidth = fm.height()
        self.setFixedWidth(preferedWidth)
        self.setMinimumHeight(preferedHeight)



from __future__ import annotations

import weakref
import logging
import numpy

from tomoscan.tomoobject import TomoObject
from tomoscan.scanbase import TomoScanBase
from tomoscan.volumebase import VolumeBase

_logger = logging.getLogger(__name__)


class StitchingMetadata:
    """
    overload of a TomoObject to register metadata set by the user on positions
    """

    KEY_PIXEL_OR_VOXEL_SIZE = "pixel_or_voxel_size"
    KEY_PIXEL_POSITION = "position_px"
    KEY_METRIC_POSITION = "position_m"

    def __init__(self, tomo_obj) -> None:
        super().__init__()
        if tomo_obj is None:
            self._tomo_obj = None
        else:
            self._tomo_obj = weakref.ref(tomo_obj)
        self._pixel_or_voxel_size = [None, None, None]
        self._pos_as_px = [None, None, None]
        self._pos_as_m = [None, None, None]

    @property
    def tomo_obj(self) -> TomoObject:
        if self._tomo_obj is None:
            return None
        else:
            return self._tomo_obj()

    def setPixelOrVoxelSize(self, value, axis):
        if value is not None and not isinstance(
            value, (float, numpy.float32, numpy.float64)
        ):
            raise TypeError(f"Invalid type for value. Got {type(value)}")
        self._pixel_or_voxel_size[axis] = value

    def setPxPos(self, value, axis):
        if value is not None and not isinstance(value, (int, numpy.int32)):
            raise TypeError(f"Invalid type for value. Got {type(value)}")
        self._pos_as_px[axis] = value

    def setMetricPos(self, value, axis):
        if value is not None and not isinstance(
            value, (int, numpy.float32, numpy.float64, float)
        ):
            raise TypeError(f"Invalid type for value. Got {type(value)}")
        self._pos_as_m[axis] = value

    def get_raw_position_m(self, axis):
        if self.tomo_obj is None:
            return
        bb = self.tomo_obj.get_bounding_box(axis)
        return (bb.max - bb.min) / 2 + bb.min

    def get_abs_position_px(self, axis, warn=True) -> int | None:
        if axis not in (0, 1, 2):
            raise ValueError("axis is expected to be in (0, 1, 2)")
        # if user registered position as pixels
        if self._pos_as_px[axis] is not None:
            return int(self._pos_as_px[axis])
        else:
            # get position as m from the user input
            pos_as_m = self._pos_as_m[axis]
            # or from metadata if not provided
            if pos_as_m is None:
                try:
                    pos_as_m = self.get_raw_position_m(axis=axis)
                except Exception as e:
                    if warn:
                        _logger.warning(
                            f"Fail to get bounding box of {self.tomo_obj}. Error is {e}"
                        )
                    return None
            # else if user registered position as meter
            pixel_or_voxel_size = self.get_pixel_or_voxel_size(axis=axis)
            if pixel_or_voxel_size is None:
                if warn:
                    if isinstance(self.tomo_obj, TomoScanBase):
                        missing = "pixel"
                    else:
                        missing = "voxel"
                    _logger.warning(
                        f"Unable to get {missing} from {self.tomo_obj}. Please provide it manually"
                    )
                return None
            else:
                return int(pos_as_m / pixel_or_voxel_size)

    def reset(self):
        self._pixel_or_voxel_size = [None, None, None]
        self._pos_as_px = [None, None, None]
        self._pos_as_m = [None, None, None]

    def get_pixel_or_voxel_size(self, axis) -> float | None:
        """
        return value provided by the user if any else return the value contained in metadata
        """
        if self.tomo_obj is None:
            return
        if self._pixel_or_voxel_size[axis] is not None:
            return self._pixel_or_voxel_size[axis]
        elif isinstance(self.tomo_obj, TomoScanBase):
            if axis == 0:
                return self.tomo_obj.sample_y_pixel_size
            elif axis in (1, 2):
                return self.tomo_obj.sample_x_pixel_size
            else:
                raise TypeError(f"axis is expected to be in (0, 1, 2). {axis} provided")
        elif isinstance(self.tomo_obj, VolumeBase):
            voxel_size = self.tomo_obj.voxel_size
            if voxel_size is None:
                return None
            else:
                return voxel_size[axis]
        else:
            raise NotImplementedError

    def to_dict(self) -> dict:
        def cast_values(values):
            return ",".join(["" if val is None else str(val) for val in values])

        return {
            self.KEY_PIXEL_OR_VOXEL_SIZE: cast_values(self._pixel_or_voxel_size),
            self.KEY_PIXEL_POSITION: cast_values(self._pos_as_px),
            self.KEY_METRIC_POSITION: cast_values(self._pos_as_m),
        }

    @staticmethod
    def from_dict(ddict, tomo_obj):
        res = StitchingMetadata(tomo_obj=tomo_obj)
        res.load_from_dict(ddict)
        return res

    def load_from_dict(self, ddict) -> dict:
        def cast_str(my_str: str, expected_type):
            return list(
                [
                    None if part_ == "" else expected_type(part_)
                    for part_ in my_str.split(",")
                ]
            )

        if self.KEY_METRIC_POSITION in ddict:
            self._pos_as_m = cast_str(
                ddict[self.KEY_METRIC_POSITION], expected_type=float
            )
        if self.KEY_PIXEL_POSITION in ddict:
            self._pos_as_px = cast_str(
                ddict[self.KEY_PIXEL_POSITION], expected_type=int
            )
        if self.KEY_PIXEL_OR_VOXEL_SIZE in ddict:
            self._pixel_or_voxel_size = cast_str(
                ddict[self.KEY_PIXEL_OR_VOXEL_SIZE], expected_type=float
            )

    def __eq__(self, __o: object) -> bool:
        return (
            self._pixel_or_voxel_size == __o._pixel_or_voxel_size
            and self._pos_as_px == __o._pos_as_px
            and self._pos_as_m == __o._pos_as_m
        )

    def __str__(self) -> str:
        metadatas = {
            "pixel_or_voxel_size": self._pixel_or_voxel_size,
            "metric position (m)": self._pos_as_m,
            "pixel position (px)": self._pos_as_px,
        }
        metadata_str = "\n -".join(
            [f"{key}: {value}" for key, value in metadatas.items()]
        )
        tomo_obj = self.tomo_obj or "?"
        return (
            f"user override stitching metadata for {tomo_obj} metadata: \n"
            + metadata_str
        )



# coding: utf-8
from __future__ import annotations


import gc
import os
import shutil
import tempfile

from orangecanvas.scheme.readwrite import literal_dumps
from silx.gui import qt
from silx.gui.utils.testutils import TestCaseQt

from orangecontrib.tomwer.widgets.visualization.DataViewerOW import DataViewerOW
from tomwer.core.utils.scanutils import MockNXtomo


class TestDataViewerOW(TestCaseQt):
    def setUp(self):
        self.tmp_dir = tempfile.mkdtemp()
        self.scan = MockNXtomo(
            scan_path=os.path.join(self.tmp_dir, "myscan"),
            n_proj=20,
            n_ini_proj=20,
            dim=10,
        ).scan
        self.widget = DataViewerOW()

    def tearDown(self):
        self.widget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self.widget.close()
        self.widget = None
        shutil.rmtree(self.tmp_dir)
        self.qapp.processEvents()
        gc.collect()

    def test(self):
        self.widget.addScan(self.scan)
        for mode in (
            "projections-radios",
            "slices",
            "raw darks",
            "raw flats",
            "reduced darks",
            "reduced flats",
        ):
            self.widget.viewer.setDisplayMode(mode)
            self.qapp.processEvents()

    def test_literal_dumps(self):
        """
        test settings are correcly dump to literals (and load)
        """
        self.widget._updateSettings()
        config = self.widget._viewer_config
        literal_dumps(config)
        self.widget._setSettings(config)



from __future__ import annotations

from importlib.metadata import version as get_version
import tomoscan.version

import ewokscore
import ewoksorange
import tomwer.version
import sluurp


try:
    import nxtomomill.version
except ImportError:
    has_nxtomomill = False
else:
    has_nxtomomill = True
try:
    import nabu
except ImportError:
    has_nabu = False
else:
    has_nabu = True
try:
    import nxtomo.version
except ImportError:
    has_nxtomo = False
else:
    has_nxtomo = True


def get_tomotools_stack_versions() -> dict[str, str]:
    """Return the version of the main libraries used by tomwer"""
    stack = {
        "tomwer": tomwer.version.version,
    }

    if has_nabu:
        stack["nabu"] = nabu.version
    if has_nxtomo:
        stack["nxtomo"] = nxtomo.version.version
    if has_nxtomomill:
        stack["nxtomomill"] = nxtomomill.version.version
    stack["tomoscan"] = tomoscan.version.version
    try:
        stack["ewokscore"] = ewokscore.__version__
    except AttributeError:
        stack["ewokscore"] = get_version("ewokscore")

    try:
        stack["ewoksorange"] = ewoksorange.__version__
    except AttributeError:
        stack["ewoksorange"] = get_version("ewoksorange")
    stack["sluurp"] = sluurp.__version__

    return stack



# coding: utf-8
from __future__ import annotations

import logging

from orangewidget import gui
from orangewidget.widget import Input, Output, OWBaseWidget
from tomoscan.series import Series

import tomwer.core.process.control.tomoobjseries
from tomwer.core.tomwer_object import TomwerObject
from tomwer.gui.control.series.seriescreator import SeriesWidgetDialog

logger = logging.getLogger(__name__)


class TomoObjSeriesOW(OWBaseWidget, openclass=True):
    name = "series of objects"
    id = "orange.widgets.tomwer.tomoobjseriesow"
    description = "Allow user define a series of object that will be defined as a Series (grouped together and can be used within a purpose like stitching)"
    icon = "icons/tomoobjseries.svg"
    priority = 55
    keywords = ["tomography", "selection", "tomwer", "series", "group"]

    ewokstaskclass = tomwer.core.process.control.tomoobjseries._TomoobjseriesPlaceHolder

    want_main_area = True
    want_control_area = False
    resizing_enabled = True

    class Inputs:
        tomo_obj = Input(name="tomo obj", type=TomwerObject, multiple=True)

    class Outputs:
        series = Output(name="series", type=Series)

    def __init__(self, parent=None):
        """ """
        super().__init__(parent)
        layout = gui.vBox(self.mainArea, self.name).layout()

        self._widget = SeriesWidgetDialog(self)
        layout.addWidget(self._widget)

        # connect signal / slot
        self._widget.sigSeriesSelected.connect(self._send_series)

    @Inputs.tomo_obj
    def addTomoObj(self, tomo_obj, *args, **kwargs):
        if tomo_obj is not None:
            self._widget.add(tomo_obj)

    def _send_series(self, series: Series):
        if not isinstance(series, Series):
            raise TypeError(
                f"series is expected to be an instance of {Series}. Not {type(series)}"
            )
        self.Outputs.series.send(series)



# coding: utf-8

import logging

_logger = logging.getLogger(__name__)


def getNumberOfDevice():
    try:
        import pycuda  # noqa F401  pylint: disable=E0401
        from pycuda import compiler  # noqa F401  pylint: disable=E0401
        from pycuda import driver  # pylint: disable=E0401

        driver.init()
        return driver.Device.count()
    except Exception:
        _logger.error("fail to discover the number of gpu")
        return None
